/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.manager.requests;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.pua.prot.manager.ScriptAnswer;
import de.bsvrz.pua.prot.manager.ScriptRelatedOperation;
import de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface;
import de.bsvrz.pua.prot.util.ProtocolRequestResult;
import de.bsvrz.sys.funclib.losb.datk.AtlExtra;
import de.bsvrz.sys.funclib.losb.datk.AtlMeta;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.Charset;

/**
 * Auftrag "Skriptquelltext zurückgeben". Siehe Feinentwurf Kapitel 2.1.2.2.1.3. Der Quelltext eines Skripts wird auf den Datenverteiler gelegt.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class RetrieveScriptSourceRequest extends ScriptRequest {

	/**
	 * @param dav         Verbindung zum Datenverteiler.
	 * @param dataManager Data Manager.
	 * @param resultData  Auftrag.
	 */
	public RetrieveScriptSourceRequest(ClientDavInterface dav, DataManagerInterface dataManager, ResultData resultData) {
		super(dav, dataManager, new ResultData[]{resultData}, false);
	}

	/**
	 * Lädt den Quelltext eines Skriptes vom Dateisystem und legt ihn im Skriptobjekt auf den Datenverteiler. Vorgehensweise: 1. Dateinname aus der Konfiguration
	 * holen 2. Quelltext vom Datenträger laden 3. Antwort-Objekt zusammenstellen 4. Auf Datenverteiler legen
	 *
	 * @return Informationen über die Auftragsausführung
	 */
	@Override
    public ProtocolRequestResult process() {
		ProtocolRequestResult result = null;
		final String fileName = dataManager.getAbsolutFileNameSource(script);
		try {
			try {
				final String source = loadSource(fileName);
				final AtlMeta meta = loadMetaInformation(dataManager.getAbsolutFileNameMeta(script));
				if(ScriptAnswer.sendSource(dav, script, client, requestId, meta, new AtlExtra(client, requestId, ScriptRelatedOperation.ACK_FLAG, source))) {
					result = new ProtocolRequestResult();
				}
				else {
					result = new ProtocolRequestResult(ErrorMessage.CAN_NOT_SEND + requestId);
				}
			}
			catch(FailureException e) {
				ScriptAnswer.sendError(dav, script, client, requestId, e.getMessage());
				result = new ProtocolRequestResult(e.getMessage());
			}
		}
		catch(FailureException e) {
			result = new ProtocolRequestResult(e.getMessage());
			debug.error(e.getMessage());
		}
		return result;
	}

	/**
	 * Liefert den Inhalt einer Datei.
	 *
	 * @param fileName Dateiname
	 *
	 * @return String mit dem Dateiinhalt
	 *
	 * @throws FailureException Falls es zu einem Fehler während des Einlesens kommt
	 */
	private String loadSource(String fileName) throws FailureException {
		FileInputStream fis = null;
		BufferedReader in = null;
		try {
			StringBuffer source = new StringBuffer();

			fis = new FileInputStream(fileName);
			in = new BufferedReader(new InputStreamReader(fis, Charset.forName("ISO-8859-1")));

			String line = null;
			do {
				line = in.readLine();
				if(line != null) source.append(line + "\n");
			}
			while(line != null & !isAborted());

			if(isAborted()) {
				return "";
			}
			else {
				return source.toString();
			}
		}
		catch(FileNotFoundException e) {
			debug.fine(ErrorMessage.FILE_NOT_FOUND + fileName + ", " + e.getMessage());
			throw new FailureException(ErrorMessage.FILE_NOT_FOUND + fileName + ", " + e.getMessage(), e, LoggerException.WARNING);
		}
		catch(IOException e) {
			debug.fine(ErrorMessage.CAN_NOT_ACCESS_FILE + fileName + ", " + e.getMessage());
			throw new FailureException(ErrorMessage.CAN_NOT_ACCESS_FILE + fileName + ", " + e.getMessage(), e, LoggerException.WARNING);
		}
		finally {
			Util.close(in);
		}
	}

	/**
	 * Lädt die Metainformationen eines Skripts aus einer Datei.
	 *
	 * @param fileName Dateiname der Metainformationen
	 *
	 * @return Metainformationen
	 *
	 * @throws FailureException Fehler beim Deserialisieren
	 */
	private AtlMeta loadMetaInformation(String fileName) throws FailureException {
		Object obj = SerializerUtil.deserialize(fileName);
		if(obj instanceof AtlMeta) {
			return (AtlMeta)obj;
		}
		else {
			throw new FailureException(ErrorMessage.UNEXPECTED_DATA_TYPE + obj.getClass().getName() + ", erwartet wurde: AtlMeta", LoggerException.WARNING);
		}
	}
}
