/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.processing;

import de.bsvrz.pua.prot.client.dataobject.PuADataState;
import de.bsvrz.pua.prot.processing.onlinebuffer.ExtResultData;
import de.bsvrz.pua.prot.processing.util.BaseDataSet;
import de.bsvrz.pua.prot.processing.util.DataInformation;
import de.bsvrz.pua.prot.util.ExpressionResult;
import de.bsvrz.pua.prot.util.ImmutableResultValue;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.pua.prot.util.ResultValue;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;

/**
 * Abstrakte Basisklassen für Klassen, die aktiv Daten sammeln.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public abstract class ValueProvider {

	/** Wert der zurückgegeben wird, wenn keine Daten vorliegen. */
	private static final ImmutableResultValue NO_RV_DATA = new ImmutableResultValue(new ExpressionResult());

	/** Markierungsinformation. */
	protected DataInformation dataInformation;

	/** Markierungsinformation des vorletzten Datensatzes, der zum auffüllen verwendet wird. Markierungsinformation von {@link #_oldValue}. */
	private DataInformation _oldDataInformation;

	/** Letzter verwendbarer Datensatz zum Auffüllen. D.h. ein Archivdatensatz, der zum Auffüllen verwendet werden kann. */
	private ExtResultData _oldValue = null;

	/** Zuletzt erhaltener Datensatz */
	protected ExtResultData value = null;

	/**
	 * Hat das Objekt den kleinsten Zeitstempel. <code>true</code> -> verwende aktuellen Wert ({@link #value}). <code>false</code> -> verwende aufgeüllten Wert
	 * ({@link #_oldValue}).
	 */
	private boolean _useNewData = false;

	/**
	 * Zeigt ob der Wertebeschaffer zu einem Element gehört, das im Spalten-Bereich verwendet wird. <code>true</code>Element wird im Spalten Bereich verwendet.
	 * <code>false</code> sonst.
	 */
	private boolean _isShown;

	/** Soll ein Änderungsprotokoll erzeugt werden? */
	private ProtocolType _protocolType;

	/** <code>true</code> falls Daten aufgefüllt werden dürfen. */
	private boolean _padData;

	/**
	 * @param protocolType  Protokollart
	 * @param isShown       Wird das Element im Spalten-Bereich verwendet? <code>true</code>Element wird im Spalten Bereich verwendet.
	 * @param padData       <code>true</code> falls Daten aufgefüllt werden dürfen.
	 */
	public ValueProvider(ProtocolType protocolType, boolean isShown, boolean padData) {
		this._protocolType = protocolType;
		this._padData = padData;
		reset();
		this._isShown = isShown;
	}

	/**
	 * Erzeugt die Markierungsinformation.
	 *
	 * @return <code>true</code>: Es gibt einen Sprung im Datensatzindex. (Ist dies der Fall, muss muss ein leerer Datensatz eingefügt werden! Siehe TAnf V.1.1 S.
	 *         33, "Markierung Pot. Datenlücke"). Falls noch keine Daten vorliegen, wird <code>false</code> zurückgegeben.
	 */
	abstract public boolean updateDataInformation();

	/**
	 * Zeigt an ob noch weitere Daten geliefert werden. Der Wert wird nur durch einen Aufruf von {@link #take(long)} aktualisiert. <code>false</code> wird erst
	 * zurückgeliefert, wenn der Versuch weitere Daten mittels {@link #take(long)} abzurufen fehlschlägt.
	 *
	 * @return <code>true</code> falls noch weitere Daten geliefert werden.<code>false</code> sonst.
	 */
	abstract public boolean hasNext();

	/**
	 * Holt den nächsten Datensatz und legt ihn in {@link #value} ab.
	 *
	 * @param timeout Maximale Wartezeit. Wird die Wartezeit überschritten, so wird due Methode abgebrochen und <code>false</code> zurückgegeben. Ist dieser
	 *                Parameter <code>0</code> so wird solange gewartet, bis Daten eintreffen.
	 *
	 * @return True falls Daten empfangen wurde, false falls es innerhalb der vorgegebenen Wartezeit keine Daten gibt.
	 *
	 * @throws FailureException     Fehler beim Datenempfang.
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	abstract protected boolean executeTake(long timeout) throws FailureException, InterruptedException;

	/**
	 * Holt den nächsten Datensatz. Wartet, bis Daten empfangen werden.
	 *
	 * @param timeout Maximale Wartezeit. Wird die Wartezeit überschritten, so wird due Methode abgebrochen und <code>false</code> zurückgegeben. Ist dieser
	 *                Parameter <code>0</code> so wird solange gewartet, bis Daten eintreffen.
	 *
	 * @return True falls Daten empfangen wurde, false falls es innerhalb der vorgegebenen Wartezeit keine Daten gibt.
	 *
	 * @throws FailureException     Fehler beim Datenempfang.
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	public boolean take(long timeout) throws FailureException, InterruptedException {
		if(value != null) {
			_oldValue = value;
			_oldDataInformation = dataInformation;
		}
		return executeTake(timeout);
	}

	/**
	 * Liefert den Wert des Elements. Je nach Ergebnis von {@link #isWinner()} wird der aktuellste Wert oder ein aufgefüllter Wert zurückgegeben.
	 *
	 * @return Wert des Elements.
	 *
	 * @see BaseDataSet#isPaddedData()
	 */
	public BaseDataSet getValue() {
		if(isWinner() && value != null) {
			ResultValue rv = new ResultValue(value.getData());
			return new BaseDataSet(
					rv,
					PuADataState.of(value.getDataType()),
					dataInformation,
					value.getDataTime(),
					value.getArchiveTime(),
					BaseDataSet.INVALID_VALUE,
					BaseDataSet.INVALID_VALUE,
					value.getDataIndex(),
					false
			);
		}
		else if(_oldValue != null && _oldDataInformation.dataForPadding() && _padData) {
			return new BaseDataSet(
					new ResultValue(_oldValue.getData()),
					PuADataState.of(_oldValue.getDataType()),
					_oldDataInformation,
					_oldValue.getDataTime(),
					_oldValue.getArchiveTime(),
					BaseDataSet.INVALID_VALUE,
					BaseDataSet.INVALID_VALUE,
					_oldValue.getDataIndex(),
					true
			);
		}
		else
		{
			// Leerer Datensatz, wenn nicht aufgefüllt wird
			return new BaseDataSet(
					NO_RV_DATA,
					PuADataState.EMPTY,
					_oldDataInformation,
					BaseDataSet.INVALID_VALUE,
					BaseDataSet.INVALID_VALUE,
					BaseDataSet.INVALID_VALUE,
					BaseDataSet.INVALID_VALUE,
					BaseDataSet.INVALID_VALUE,
					true
			);
		}
	}

	/**
	 * Liefert die Markierungsinformation zurück
	 *
	 * @return Markierungsinformation
	 */
	public DataInformation getDataInformation() {
		return dataInformation;
	}


	/** Bricht das Sammeln der Daten ab. */
	abstract public void abort();


	/**
	 * Dieses Objekt hat beim Verschränken nicht den kleinsten Zeitstempel. Der nächste Aufruf von {@link #getValue()} muss den vorletzten, zum Auffüllen zu
	 * verwendenden Datensatz zurückgeben, und nicht den aktuellen.
	 */
	public void lost() {
		_useNewData = false;
	}

	/**
	 * Dieses Objekt hat beim Verschränken den kleinsten Zeitstempel. Falls der Wert des Elements zum Auffüllen tauglich ist, wird dieser gespeichert. Sonst wird
	 * er verworfen, d.h. es existiert kein auffüllbarer Datensatz für dieses Element mehr.
	 *
	 * @see DataInformation#dataForPadding()
	 */
	public void won() {
		_useNewData = true;
	}

	/** @return <code>true</code> falls das Element einen minimalen Zeitstempel besitzt. <code>false</code> sonst. */
	public boolean isWinner() {
		return _useNewData;
	}

	/**
	 * Liefert den Datenzeitstempel des zuletzt erhaltenen Datensatzes ab.
	 *
	 * @return Zeitstempel. Wurde kein Datensatz erhalten, so wird {@link Long#MAX_VALUE} zurückgegeben
	 */
	public long getMostRecentTimeStamp() {
		if(value != null) {
			return value.getDataTime();
		}
		else {
			return Long.MAX_VALUE;
		}
	}

	/**
	 * Liefert den Datenindex des zuletzt erhaltenen Datensatzes ab.
	 *
	 * @return Zeitstempel. Wurde kein Datensatz erhalten, so wird der Wert <code>-1</code> zurückgegeben.
	 */
	public long getMostRecentDataIndex() {
		if(value != null) {
			return value.getDataIndex();
		}
		else {
			return -1;
		}
	}

	/**
	 * Liefert den Datenindex des momentan verwendeten Datensatzes ab, der beim Aufruf von {@link #getValue()} zurückgegeben wird.
	 *
	 * @return Zeitstempel. Wurde kein Datensatz erhalten, so wird {@link Long#MAX_VALUE} zurückgegeben
	 */
	public long getUsedTimeStamp() {
		ExtResultData temp;
		if(isWinner()) {
			temp = value;
		}
		else {
			temp = _oldValue;
		}

		if(temp != null) {
			return temp.getDataTime();
		}
		else {
			return Long.MAX_VALUE;
		}
	}

	/**
	 * Liefert den Datenindex des momentan verwendeten Datensatzes ab, der beim Aufruf von {@link #getValue()} zurückgegeben wird.
	 *
	 * @return Zeitstempel. Wurde kein Datensatz erhalten, so wird der Wert <code>-1</code> zurückgegeben.
	 */
	public long getUsedDataIndex() {
		ExtResultData temp;
		if(isWinner()) {
			temp = value;
		}
		else {
			temp = _oldValue;
		}

		if(temp != null) {
			return temp.getDataIndex();
		}
		else {
			return -1;
		}
	}

	/**
	 * Zeigt ob der Wertebeschaffer zu einem Element gehört, das im Spalten-Bereich verwendet wird.
	 *
	 * @return	<code>true</code>Element wird im Spalten Bereich verwendet. <code>false</code> sonst.
	 */
	public boolean isShown() {
		return _isShown;
	}

	/**
	 * Setzt das Objekt auf den Anfangszustand zurück. Methode sollte zu Beginn eines neuen Intervalls aufgerufen werden, damit die Markierungsinformation und die
	 * Daten zum Auffüllen zurückgesetzt werden.
	 */
	protected void reset() {
		dataInformation = new DataInformation(_protocolType);
		_oldDataInformation = new DataInformation(_protocolType);
		_oldValue = null;
		value = null;
		_useNewData = false;
	}
}
