/*
 * Copyright 2015 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.sender;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

/**
 * Klasse, die der PuA-Server zur Kommunikation mit den Clients benutzt. Früher wurde der SimpleSender benutzt.
 * Details zur Notwendigkeit siehe Nerz Fehlereintrag Nerz-F-157.
 *
 * Das Verhalten der Klasse ist wie folgt:
 * <ol>
 *     <li>Eine Instanz der Klasse wird erzeugt, dadurch wird die Datenidentifikation zum Senden mit Sendesteuerugn angemeldet</li>
 *     <li>Bevor eine positive Sendesteuerugn eintritt werden mit send() übergebene Telegramme zwischengespeichert</li>
 *     <li>Sobald es eine positive Sendesteuerung gibt, werden zwischengespeicherte Telegramme gesendet</li>
 *     <li>Weitere über send() übergebene Telegramme werden direkt gesendet</li>
 *     <li>Sobald eine negative Sendesteuerung eintrifft oder wenn es nach 1 Minute (oder einem anderen angegebenen Timeout) keine positive Sendesteuerung gab,
 *     wird der Sender abgemeldet und die Aufrufende Klasse über einen callback informiert, dass der Client nicht mehr gültig ist.
 *     Dadurch kann bspw. der Client aus datenstrukturen entfernt werden oder ein evtl. gerade durchgeführte Protokollerstellung
 *     abgebrochen werden</li>
 * </ol>
 *
 * Sämtliche send()-Aufrufe sind nicht blockierend, es wird also nicht gewartet, bis die Daten wirklich versand werden konnten.
 * Die Sendeanmeldung wird erst entfernt, wenn sich der Client als Empfänger abgemeldet oder aus anderen Gründen beendet hat.
 *
 * @author Kappich Systemberatung
 */
public class PuaSender {

	// Noch keine Sendesteuerung
	public static final byte NO_SEND_CONTROL_YET = -1;

	// Keine positive Sendesteuerung nach 1 Minute
	public static final byte TIMEOUT = -2;

	private final ClientSenderInterface _mySender = new ClientSenderInterface() {
		@Override
		public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
			// updateState wird hier vom _sendControlService ausgeführt um den Sendesteuerungs-Thread nicht zu blockieren
			// und alle Ausführungen von updateState im selben Thread zu halten.
			_sendControlService.submit((Runnable)()->updateState(state));
		}

		@Override
		public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
			return true;
		}
	};

	// Verbindung
	private final ClientDavInterface _dav;

	// objekt für das Daten gesendet werden sollen
	private final SystemObject _object;

	// dataDescription für die Daten gesendet werden sollen
	private final DataDescription _dataDescription;

	// Callback der aufgerufen wird, wenn sich der Client abgemeldet hat oder ein Timeout auftrat (also wenn die Verbindung zum Client unterbrochen wurde)
	private final Runnable _closedCallback;

	// Zustand Sendesteuerung
	private volatile byte _state = NO_SEND_CONTROL_YET;

	// Queue für noch nicht gesendete Daten (der Anwender der Klasse stellt über Flusskontrolle o.ä. sicher, dass diese nicht zu voll läuft)
	private final Deque<Data> _queue = new ArrayDeque<>();

	// Thread, der Sendesteuerungen und Timeouts behandelt
	private static final ScheduledExecutorService _sendControlService = Executors.newSingleThreadScheduledExecutor();

	// Future, was den Timeout abbildet (wird gespeichert, damit der Timer abgebrochen werden kann sobald eine positive Sendesteuerung eintrifft)
	private final ScheduledFuture<?> _timeoutFuture;

	private static final Debug _debug = Debug.getLogger();

	private volatile boolean _isAlive = true;
	
	private final Object _flushLock = new Object();
	
	// Wenn ein Datensatz nicht gesendet wurde, damit nicht davon ausgegeangen wird, dass alles okay ist
	private boolean _allDataSend = true;

	public PuaSender(ClientDavInterface dav, SystemObject object, DataDescription dataDescription, Runnable closedCallback) {
		this(dav, object, dataDescription, closedCallback, 60000);
	}

	public PuaSender(ClientDavInterface dav, SystemObject object, DataDescription dataDescription, Runnable closedCallback, long timeout) {
		_dav = dav;
		_object = object;
		_dataDescription = dataDescription;
		_closedCallback = closedCallback;
		_timeoutFuture = _sendControlService.schedule((Runnable) () -> updateState(TIMEOUT), timeout, TimeUnit.MILLISECONDS);
		try {
			ConnectionManager.subscribeSender(dav, _mySender, object, dataDescription, SenderRole.sender());
		}
		catch(OneSubscriptionPerSendData cause) {
			// Dies sollte nicht passieren, da wir die einzige Klasse sind, die das _mySender-Objekt kennt
			_debug.error("Fehler bei der Anmeldung als Sender", cause);
		}
	}

	public void send(Data data) {
		try {
			if(!_isAlive) {
				_allDataSend = false;
				return;
			}
			if(_state == ClientSenderInterface.START_SENDING) {
				_dav.sendData(new ResultData(_object, _dataDescription, _dav.getTime(), data));
			}
			else {
				synchronized(_queue) {
					_queue.addLast(data);
				}
			}
		}
		catch(SendSubscriptionNotConfirmed | DataNotSubscribedException cause) {
			// das kann passieren wenn gerade Daten gesendet werden sollen und der Client plötzlich weg ist.
			// In dem Fall können die Daten verworfen werden.
			_debug.fine("Fehler beim Senden von Daten", cause);
			_allDataSend = false;
			updateState(ClientSenderInterface.STOP_SENDING);
		}
	}

	private void updateState(final byte state) {
		byte oldState = _state;
		if(state == oldState || !isAlive()) {
			return;
		}
		if(state == TIMEOUT && oldState == ClientSenderInterface.START_SENDING) {
			/* Sollte der Timeout beim Bestehen einer positiven Sendesteuerung eintreffen 
			(obwohl der Future gecancelt wurde), nichts tun */
			return;
		}
		_state = state;
		if(_state == ClientSenderInterface.START_SENDING) {
			// Empfänger ist jetzt vorhanden, ggf. gecachte Daten senden
			_timeoutFuture.cancel(false);
			while(queueIsEmpty()) {
				Data data;
				synchronized(_queue) {
					data = _queue.removeFirst();
				}
				send(data);
			}
		}
		else if(oldState == ClientSenderInterface.START_SENDING) {
			// Empfänger ist "verschwunden"
			_isAlive = false;
			ConnectionManager.unsubscribeSender(_dav, _mySender, _object, _dataDescription);
			synchronized(_queue) {
				if(!_queue.isEmpty()) {
					_allDataSend = false;
					_queue.clear();
				}
				_queue.notifyAll();
			}
			_closedCallback.run();
		}
	}

	private boolean queueIsEmpty() {
		synchronized(_queue) {
			return !_queue.isEmpty();
		}
	}

	@Override
	public String toString() {
		return _object.getPidOrId() + ":" + _dataDescription.getAttributeGroup().getPidOrId() + ":" + _dataDescription.getAspect().getPidOrId();
	}

	public SystemObject getObject() {
		return _object;
	}

	public DataDescription getDataDescription() {
		return _dataDescription;
	}

	public ClientDavInterface getDav() {
		return _dav;
	}

	public boolean isAlive() {
		return _isAlive;
	}

	/**
	 * Wartet, bis alle verbleibenden Datensätze gesendet wurden
	 *
	 * @return true wenn alle bisher mit send() übergebenen Datensätze erfolgreich gesendet wurden, false wenn es nach dem Timeout keine positive Sendesteuerung
	 * gibt oder sich der Empfänger abgemeldet hat bevor alle Daten gesendet werden konnten oder der Thread interrupted wurde.
	 */
	public boolean flush() {
		try {
			synchronized(_queue) {
				while(_isAlive && !_queue.isEmpty()) {
					_queue.wait();
				}
			}
		}
		catch(InterruptedException ignored){
			return false;
		}
		return _queue.isEmpty() && _allDataSend;
	}
}
