/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.manager.requests;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.pua.prot.manager.ProtocolAnswer;
import de.bsvrz.pua.prot.manager.datamanager.DataManager;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.pua.prot.util.ProtocolRequestResult;
import de.bsvrz.sys.funclib.losb.datk.AtgProtocolRequest;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;

/**
 * Basisklasse für die Anfragen zum abrufen von Protokolllisten.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ProtocolListRequest extends ProtocolRequest {

	private final String _client;

	/** Auftragsparameter */
	private ResultData _resultData;

	/** Dateiendung der Protokolldatei */
	private String _extension;

	/**
	 * Erzeugt das Objekt.
	 *
	 * @param dav         Verbindung zum Datenverteiler
	 * @param data        Data von resultData
	 * @param resultData  Auftragsparameter. resultData.getData() darf nicht null liefern!
	 * @param dataManager Data Manager
	 * @param extension   Dateiendung der Protokolldatei
	 */
	public ProtocolListRequest(final ClientDavInterface dav, final Data data, final ResultData resultData, final DataManager dataManager, final String extension) {
		super(dav, new ResultData[]{resultData}, dataManager);
		final AtgProtocolRequest request = AtgProtocolRequest.getJavaObject(data);

		_client = (String)SerializerUtil.deserialize(request.requestData);

		this._resultData = resultData;
		this._extension = extension;
	}

	/**
	 * Erstellt eine Liste aller vorhandener Protokoll-Dateien und zusätzlicher Informationen und sendet sie an den Auftraggeber. (siehe Feinentwurf, Kapitel
	 * 2.1.2.2.2.4)
	 *
	 * @see de.bsvrz.pua.prot.manager.requests.DataManagerRequest#process()
	 */
	@Override
    public ProtocolRequestResult process() {
		//Liste mit Dateien holen
		final File dir = new File(dataManager.getPath());
		final File[] protocols = getPossibleProtocols(dir);

		//Protokoll-Ids & Parameter lesen
		long protocolIds[] = new long[protocols.length];
		ProcessingParameter pp[] = new ProcessingParameter[protocolIds.length];
		int i = 0;
		for(final File file : protocols) {
			if(isAborted()) break;
			try {
				final Long protocolId = new Long(file.getName().replace(_extension, ""));
				if(protocolId != null) {
					protocolIds[i] = protocolId;
					final ProcessingParameter p;
					try {
						p = getParameters(file);
					}
					catch(FailureException e) {
						debug.warning("Kann gespeichertes Protokoll nicht lesen", e);
						continue;
					}
					// Private Protokolle des nicht aktuellen Benutzers ignorieren.
					if(p.isPublic() || p.getCreatorName() == null || p.getCreatorName().equals("") || p.getCreatorName().equals(this._client)) {
						pp[i] = p;
						i++;
					}
				}
			}
			catch(NumberFormatException e) {
				debug.fine(ErrorMessage.CAN_NOT_GET_PROTOCOL_ID + file.getName() + ", " + e.getMessage());
				return new ProtocolRequestResult(ErrorMessage.CAN_NOT_GET_PROTOCOL_ID + file.getName() + ", " + e.getMessage());
			}
			catch(IOException e) {
				debug.error(ErrorMessage.CAN_NOT_ACCESS_FILE + file, e);
				return new ProtocolRequestResult(ErrorMessage.CAN_NOT_ACCESS_FILE + file + ": " + e.getMessage());
			}
		}
		if(!isAborted()) {
			if(i < protocolIds.length)	// Größe des Array anpassen.
			{
				final long[] temp = new long[i];
				final ProcessingParameter[] tempPP = new ProcessingParameter[i];
				System.arraycopy(protocolIds, 0, temp, 0, temp.length);
				protocolIds = temp;
				System.arraycopy(pp, 0, tempPP, 0, tempPP.length);
				pp = tempPP;
			}
			return sendAnswer(protocolIds, pp);
		}
		else {
			return new ProtocolRequestResult();
		}
	}

	private ProtocolRequestResult sendAnswer(final long[] protocolIds, final ProcessingParameter[] pp) {
		final AtgProtocolRequest request = AtgProtocolRequest.getJavaObject(_resultData.getData());

		//Antwort senden
		try {
			ProtocolAnswer.sendProtocolList(ProtocolAnswer.sender(dav, request.client), request.requestId, protocolIds, pp);
			return new ProtocolRequestResult();
		}
		catch(FailureException e) {
			debug.error(e.getMessage());
			return new ProtocolRequestResult(e.getMessage());
		}
	}

	/**
	 * Liefert alle Dateien mit der Dateinamenserweiterung DataManager.PROTOCOL_FILENAME_EXTENSION
	 *
	 * @param dir Verzeichnis, das durchsucht wird.
	 *
	 * @return Liste mit allen gefundenen Dateien.
	 */
	private File[] getPossibleProtocols(final File dir) {
		return dir.listFiles(
				new FileFilter() {
					public boolean accept(final File file) {
						if(file.isFile() && file.getName().endsWith(_extension)) {
							return true;
						}
						else {
							return false;
						}
					}
				}
		);
	}

	/**
	 * Liest Auftragsparameter der Protokollerstellung aus Datei.
	 *
	 * @param file Datei.
	 *
	 * @return Auftragsparameter.
	 *
	 * @throws IOException      Fehler beim Dateizugriff.
	 * @throws FailureException Keine ProcessingParameter vorhanden.
	 */
	private ProcessingParameter getParameters(final File file) throws IOException, FailureException {
		final ObjectInputStream ois;
		ois = new ObjectInputStream(new FileInputStream(file));
		try {
			final Object o;
			try {
				o = ois.readObject();
			}
			catch(ClassNotFoundException e) {
				throw new FailureException(e, LoggerException.WARNING);
			}
			if(o instanceof byte[]) {
				return ProcessingParameter.deserialize(dav.getDataModel(), (byte[])o);
			}
			else {
				throw new FailureException(
						ErrorMessage.UNEXPECTED_DATA_TYPE + o.getClass().getName() + " byte[] in GetAllProtocolRequest.getParameters.", LoggerException.WARNING
				);
			}
		}
		finally {
			ois.close();
		}
	}
}
