/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.processing;

import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.pua.prot.client.dataobject.*;
import de.bsvrz.pua.prot.exceptions.NoncriticalException;
import de.bsvrz.pua.prot.manager.ProtocolAnswer;
import de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface;
import de.bsvrz.pua.prot.manager.datamanager.ProtocolOutputStream;
import de.bsvrz.pua.prot.processing.archivebuffer.ArchiveBuffer;
import de.bsvrz.pua.prot.processing.onlinebuffer.OnlineBuffer;
import de.bsvrz.pua.prot.processing.util.BaseDataSet;
import de.bsvrz.pua.prot.processing.util.DataInformation;
import de.bsvrz.pua.prot.processing.util.IntermediateDataSet;
import de.bsvrz.pua.prot.sender.PuaSender;
import de.bsvrz.pua.prot.util.*;
import de.bsvrz.pua.prot.util.ProcessingInformation.ApplyAggregation;
import de.bsvrz.pua.prot.util.attributes.*;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.AtlMeta;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;
import de.bsvrz.sys.funclib.losb.util.Tuple;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Führt die Datenaufbereitung durch.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class Processing extends BasicFlowControl implements ProcessingInterface {

	/**
	 * Falls der Online Buffer mehr Ausgangsdatensätze bereithält als <code>SUSPEND_ARCHIVE</code> wird die Verarbeitung von Archivdaten unterbrochen, um die
	 * Ausgangsdatensätze vom OnlineBuffer abzusenden.
	 */
	private static final long SUSPEND_ARCHIVE = 5000;

	/** Debug-Ausgaben */
	private static final Debug _debug = Debug.getLogger();

	/** Sammelt dirket vom Dav empfangene Online-Daten. */
	private OnlineBuffer _onlineBuffer = null;

	/** Fordert Daten vom Archivsystem an. */
	private ArchiveBuffer _archiveBuffer = null;

	/** Protokoll-Id */
	private final long _protocolId;

	/** Informationen zur Datenaufbereitung. */
	private final ProcessingInformation _pi;

	/** Objekt zum Versand der Daten */
	private final PuaSender _sender;

	/** Metainformationen zum Skript. Werden für den Administrationsdatensatz benötigt. */
	private final AtlMeta _atlMeta;

	/** Ende der letzten Periode. */
	private final long _finalPeriod;

	/** Auftragsparameter */
	private final ProcessingParameter _pp;

	/** Verbindung zum Daten Manager. Bei Bedarf speichert Daten Manager das Protokoll. */
	private final DataManagerInterface _dataManager;

	/** Zeigt an ob das Protokoll gespeichert werden soll. Kommt es beim Speichern zu einem Fehler, wird diese Variable auf false gesetzt. */
	private boolean _saveProtocol = false;

	/** Zeigt ob die Protokollerstellung abgebrochen wurde. */
	private volatile boolean _abort = false;

	/** Konfigurationsverantwortlicher für Anfragen an das Archivsystem. */
	private final ConfigurationObject _configAuth;

	/**
	 * Erlaubt schnellen Zugriff auf die unterschiedlichen Datenidentifikationen, die als Spalten Skript verwendet werden. Jeder Eintrag ist ein Index für
	 * BaseDataSet.
	 */
	private int[] _dataIdentifications = null;

	/** Status der Datenaufbereitung */
	private byte _status = 0;

	/** Objekt zum Speichern des Protokolls. */
	private ProtocolOutputStream _protocolOutput = null;

	/** Objekt zum Zippen und Serialisieren der Ergebnisdaten. INFO ggf. in den Sender verschieben. */
	private final SerializerUtil _serializer;

	/** Sollen die Ergebnisdatensätze auf den Dav gelegt werden? Wenn nein, dann wird das Protokoll nur gespeichert und kann ggf. später abgerufen werden. */
	private boolean _send = true;

	/** Zeigt an, dass ein Sendefehler bereits geloggt wurde. */
	private boolean sendFailureWasLogged;

	/** Der Thread der den Vorgang ausführt. Wird für interrupt() benutzt um den Vorgang abzubrechen. */
	private Thread _thread;

	private JobState _jobState = JobState.InProgress;

	/**
	 * Erzeugt ein Objekt zur Datenaufbereitung.
	 * @param sender      Objekt zum Versand der Daten.
	 * @param dataManager Verbindung zum Daten Manager. Bei Bedarf speichert Daten Manager das Protokoll.
	 * @param serializer  Objekt zum Zippen und Serialisieren der Ergebnisdaten.
	 * @param protocolId  Id des zu erstellenden Protokolls.
	 * @param pp          Auftragsparameter.
	 * @param pi          Zur Protokollerstellung benötigte Informationen.
	 * @param atlMeta     Metainformationen zum Skript aus dem <code>pi</code> generiert wurde. Werden für den Administrationsdatensatz benötigt.
	 * @param configAuth  Konfigurationsverantwortlicher für Anfragen an das Archivsystem.
	 * @param maxDelay    Maximale Zeitdauer in ms für die die Protokollerstellung angehalten werden kann.
	 * @param timeout Timeout
	 * @param initialFlowControl InitialFlowControl
	 */
	public Processing(
			final PuaSender sender,
			final DataManagerInterface dataManager,
			final SerializerUtil serializer,
			final long protocolId,
			final ProcessingParameter pp,
			final ProcessingInformation pi,
			final AtlMeta atlMeta,
			final ConfigurationObject configAuth,
			final long maxDelay,
			final int timeout, final long initialFlowControl) {
		super(maxDelay, timeout, initialFlowControl);
		this._sender = sender;
		this._dataManager = dataManager;
		this._serializer = serializer;
		this._protocolId = protocolId;
		this._pi = pi;
		this._pp = pp;
		this._atlMeta = atlMeta;
		this._configAuth = configAuth;
		_finalPeriod = pi.getPeriods().get(pi.getPeriods().size() - 1).last;
		_thread = Thread.currentThread();
	}

	/** @see de.bsvrz.pua.prot.processing.ProcessingInterface#process(boolean) */
	public void process(final boolean abortProcessing) {
		this._abort = abortProcessing;
		ProtocolRequestResult result = null;
		String exception = null;
		try {
			// Daten vom Ars & Dav sammeln. Methoden kehren sofort zurück.
			if(!_abort) {
				final long stopCollectingData = collectDataFromDav();
				collectDataFromArs(stopCollectingData);
			}

			// ggf Protokoll speichern:
			prepareProtocolWriter();

			// Administrations & Aggregationsdatensatz zusammenbauen & versenden
			sendAdministrationInformation();
			sendAggregationInformation();

			// Ars- & Dav-Daten verarbeiten
			IntermediateDataSet data; // Ausgangsdatensatz, Status
			// Schleife läuft bis: Abbruch oder keine Daten mehr von den Buffern
			final long timestamp = System.currentTimeMillis();
			int dataNumber = 0;
			while(!_abort && (!_archiveBuffer.isDone() || !_onlineBuffer.isDone())) {
				if(_send) flowControl();

				/*
				 * Erst alle Online-Daten abarbeiten: ArS hat ja eine eingebaute Flußkontrolle -> ArchiveBuffer kann
				 * nicht überlaufen. Der Online-Buffer schon!
				 */
				while(!_abort && _onlineBuffer.hasData() && !_pause.get()) {
					dataNumber++;
					data = _onlineBuffer.take();
					sendProtocolResultData(data);
				}

				while(!_abort && _archiveBuffer.hasData() && _onlineBuffer.size() < SUSPEND_ARCHIVE && !_pause.get()) {
					dataNumber++;
					data = _archiveBuffer.take();
					sendProtocolResultData(data);
				}

				synchronized(this) {
					/*
					 * archiveBuffer weckt Processing nachdem alle Daten erhalten wurden. onlineBuffer wartet, bis Daten
					 * eintreffen. Wenn keine Daten eintreffen, dann wartet onlineBuffer ewig, und wekct Processing
					 * nicht mehr auf. -> Processing muss selber sicherstellen, dass es wieder aufwacht nextIntervall
					 * stellt ggf. Archivanfrage für nächstes Intervall
					 */
					if(!_abort && !_archiveBuffer.hasData()) // 1. !abort ohne diese Prüfung funktioniert das abbrechen
					// nicht: Das notifyAll aus abort() wird aufgerufen, bevor
					// das wait() Statement erreicht ist -> Thread wartet,
					// obwohl er abgebrochen werden soll.
					{ // 2. !archieBuffer.hasData(): Es ist möglich, dass Archivpuffer Schleife beendet wird, obwohl
						// Daten im Archivsystem vorliegen (z.B. weil zu viele Daten im OnlinePuffer sind. Der
						// Archivbuffer läuft über, und legt sich schlafen, bis ein 'take()' aufgerufen wird. Ohne diese
						// Überprüfung läuft Processing dann in das wait() und wartet ewig - da der Archivbuffer je erst
						// beim nächsten 'take()' geweckt wird.
						if(_archiveBuffer.isDone()) {
							wait(getMaxWait()); // Archiv ist fertig -> es wird nur noch auf Online-Daten gewartet
						}
						else {
							wait(); // Archiv ist noch nicht fertig -> wird Processing aufwecken
						}
					}
				}
			}

			final double duration = (System.currentTimeMillis() - timestamp) / 1000.0;
			_debug.info(
					String.format("Die Datenaufbereitung für Protokoll %d ist abgeschlossen. Es wurden %d Datensätze versandt.\nPerformanceInfo Gesamt: %d Datensätze, %.1fs => %.3f Datensätze/s", _protocolId, dataNumber, dataNumber, duration, dataNumber / duration)
			);
		}
		catch(InterruptedException e) { // exception wird nicht gesetzt, da es sich nicht um einen Fehler handelt (InterruptedException wird
			// geworfen, wenn die DA abgebrochen werden soll.
			_debug.finest(ErrorMessage.INTERRUPTED_EXCEPTION + _protocolId);
			_abort = true;
		}
		catch(FailureException e) {
			e.printStackTrace();
			exception = e.getMessage();
		}
		catch(Exception e) {
			exception = Util.getStackTrace(e);
		}
		finally {
			_debug.fine("Processing hat die Datensammlung abgeschlossen...räume auf & sende abschließenden Datensatz");
			cleanUp(exception);
		}
	}

	/**
	 * Sendet abschließenden Datensatz. Bricht falls nötig den {@link #_archiveBuffer} und {@link #_onlineBuffer} ab.
	 *
	 * @param exception Fehlermeldung oder null, falls kein Fehler aufgetreten ist.
	 *
	 * @return Informationen über den Ablauf der Operation.
	 */
	private ProtocolRequestResult cleanUp(final String exception) {
		if(exception != null || _abort || !_archiveBuffer.getResult().isRequestSuccessful() || !_onlineBuffer.getResult().isRequestSuccessful()) {
			if(_archiveBuffer != null) _archiveBuffer.abort();
			if(_onlineBuffer != null) _onlineBuffer.abort();

			this.abort();
		}
		// Abschließenden Ende-Datensatz senden:
		try {
			if(!_abort) {
				_debug.finest("Sende Endedatensatz für " + _protocolId);
				if(_send) {
					tryToSend(ProtocolAnswer.sendAck(_sender, _protocolId));
					if(!_sender.flush()) _send = false;
				}
			}
			else {

				String errorMsg = "";
				if(exception != null) errorMsg += exception + " ";
				if(_onlineBuffer != null && _onlineBuffer.getResult() != null && !_onlineBuffer.getResult().isRequestSuccessful()) {
					errorMsg += _onlineBuffer.getResult().getErrorMsg() + " ";
				}
				if(_archiveBuffer != null && _archiveBuffer.getResult() != null && !_archiveBuffer.getResult().isRequestSuccessful()) {
					errorMsg += _archiveBuffer.getResult().getErrorMsg() + " ";
				}
				if(errorMsg.isEmpty()){
					// Abbruch durch Benutzer
					errorMsg = ProcessingErrorMessage.ABORT_PROTOCOL;
				}
				_debug.finest("Protokollerstellung wird abgebrochen :" + _protocolId + ", Grund: " + errorMsg);
				if(_send) tryToSend(ProtocolAnswer.sendAbort(_sender, _protocolId, errorMsg));
			}
			closeProtocolWriter();
		}
		catch(FailureException e) {
			e.log();
		}

		if(exception != null) {
			_debug.warning(exception);
			return new ProtocolRequestResult(exception);
		}
		else {
			return new ProtocolRequestResult();
		}
	}

	private void tryToSend(final boolean sendOK) {
		if(sendOK){
			datasetSend();
		}
		this._send = sendOK;
		if(!_send && !sendFailureWasLogged) {
			sendFailureWasLogged = true;
			_debug.warning("Ergebnisdatensatz wurde nicht gesendet: " + _protocolId);
		}
	}

	/**
	 * @return Zeit in Millisekunden, bis das Ende des letzten Zeitbereiches erreicht wird. Falls dieser Wert negativ oder 0 ist, wird 1 zurückgeliefert. (Sonst
	 *         würde <code>wait(0)</code> ewig warten.
	 */
	private long getMaxWait() {
		final long mw = _finalPeriod - System.currentTimeMillis();
		if(mw < 1) {
			return 1;
		}
		else {
			return mw;
		}
	}

	/** Beendet das Speichern eines Protokolls. Falls die Protokollerstellung abgebrochen wurde, wird das Protokoll gelöscht. */
	private void closeProtocolWriter() {
		_jobState = JobState.Finished;
		if(_pp.getSaveProtocol() && _protocolOutput != null) {
			try {
				if(_abort) {
					_jobState = JobState.Aborted;
					_protocolOutput.abortWriting();
				}
				else if(_send) {
					_jobState = JobState.FinishedAndSaved;
					_protocolOutput.close();
				}
				else {
					_jobState = JobState.FinishedAndUnread;
					_protocolOutput.closeClientQuit();
				}
			}
			catch(NoncriticalException e) {
				e.log();
			}
			finally {
				_protocolOutput = null;
			}
		}
	}

	/** Bereitet das Speichern des Protokolles vor, sofern dies in den Auftragsparametern vorgesehen war. */
	private void prepareProtocolWriter() {
		if(_pp.getSaveProtocol()) {
			try {
				_protocolOutput = _dataManager.createProtocolStream(_protocolId, _pp);
			}
			catch(NoncriticalException e) {
				_saveProtocol = false;
				e.log();
			}
			finally {
				_saveProtocol = true;
			}
		}
		else {
			_saveProtocol = false;
		}
	}

	/**
	 * Erzeugt und sendet einen Ergebnisdatensatz an den Auftraggeber. Speichert das Protokoll bei Bedarf.
	 *
	 * @param data Ausgangsdatensatz und Status.
	 *
	 * @throws FailureException Fehler beim Erstellen des Datensatzes oder beim Senden.
	 */
	private void sendProtocolResultData(final IntermediateDataSet data) throws FailureException, InterruptedException {
		final ProtocolResult prd;
		if(data.getStatus() == IntermediateDataSet.NO_CHANGE) {
			prd = new ProtocolNoChanges(data.getWinnerTimeStamp(), data.getOrder());
		}
		else {
			_status = data.getStatus();
			prd = createProtocolResultData(data);
		}

		if(_send) {
			if(!ProtocolAnswer.sendProtocol(_sender, _protocolId, _serializer, prd)) {
				_debug.warning("Ergebnisdatensatz wurde nicht gesendet!: " + _protocolId + ", " + prd);
				_send = false;
				sendFailureWasLogged = true;
				// falls das Protokoll nicht gespeichert wird, kann die Protokollerstellung abgebrochen werden
				if(!_saveProtocol) throw new FailureException(ProcessingErrorMessage.ABORT_NO_SAVE, LoggerException.WARNING);
			}
			else {
				datasetSend();
				flowControl();
			}
		}
		saveProtocolData(prd);
	}

	/**
	 * Erzeugt Ergebnisdatensatz aus einem Ausgangsdatensatz.
	 *
	 * @param data Ausgangsdatensatz.
	 *
	 * @return Ergebnisdatensatz.
	 */
	private ProtocolResult createProtocolResultData(final IntermediateDataSet data) {
		final List<BaseDataSet> dataSets = data.getBaseDataSets();
		final ResultValue[] values = new ResultValue[dataSets.size()];
		final PuADataState[] dataStates = new PuADataState[dataSets.size()];
		for(int i = 0; i < dataSets.size(); i++) {
			final BaseDataSet baseDataSet = dataSets.get(i);
			if(baseDataSet != null) {
				values[i] = baseDataSet.getValue();
				dataStates[i] = baseDataSet.getDataState();
			} else{
				
				values[i] = new ResultValue(new ExpressionResult());
				dataStates[i] = PuADataState.COMPUTED;
			}
		}
		if(data.getOrder() != ApplyAggregation.LISTE) {
			final long[] startTimeStamps = new long[dataSets.size()];
			final long[] endTimeStamps = new long[dataSets.size()];
			for(int i = 0; i < dataSets.size(); i++) {
				startTimeStamps[i] = dataSets.get(i).getStartTimeStamp();
				endTimeStamps[i] = dataSets.get(i).getEndTimeStamp();
			}
			return new ProtocolAggregationResultData(values, startTimeStamps, endTimeStamps, data.getOrder(), data.getStatus(), data.getIntervalIndex()) ;
		}
		else {
			final DataInformation[] dataInfo = new DataInformation[_dataIdentifications.length];
			final long[] dts = new long[_dataIdentifications.length];
			final long[] ats = new long[_dataIdentifications.length];
			final long[] di = new long[_dataIdentifications.length];

			for(int i = 0; i < _dataIdentifications.length; i++) {
				final BaseDataSet baseDataSet = dataSets.get(_dataIdentifications[i]);
				if(baseDataSet != null){
					dataInfo[i] = baseDataSet.getDataInformation();
					dts[i] = baseDataSet.getDataTimeStamp();
					ats[i] = baseDataSet.getArchiveTimeStamp();
					di[i] = baseDataSet.getDataIndex();
				}
				else {
					
					final DataInformation dataInformation = new DataInformation(ProtocolType.Undefined);
					dataInfo[i] = dataInformation;
					dts[i] = -1;
					ats[i] = -1;
					di[i] = -1;
				}
			}
			return new ProtocolResultData(values, dataStates, dataInfo, dts, ats, di, data.getStatus(), data.getIntervalIndex());
		}
	}

	/**
	 * Speichert einen Datensatz, falls {@link #_saveProtocol} <code> == true</code>.
	 *
	 * @param protocolResult Datensatz.
	 */
	private void saveProtocolData(final ProtocolResult protocolResult) {
		if(_saveProtocol) {
			try {
				_protocolOutput.appendData(protocolResult);
			}
			catch(NoncriticalException e) {
				_saveProtocol = false;
				e.log();
			}
		}
	}

	/**
	 * Stellt den Aggregationsdatensatz zusammen und sendet ihn an den Auftraggeber. Diese Methode darf erst NACH Starten des Archivbuffers aufgerufen werden, da
	 * sonst der Aggregator nicht verwendet werden kann. Diese Metode
	 *
	 * @throws FailureException Kommunikationsfehler.
	 */
	private void sendAggregationInformation() throws FailureException {
		final List<Elements> attributes = _pi.getAttributes();
		List<ApplyAggregation> aggregationList = _pi.getOrder();
		if(aggregationList.size() == 0){
			aggregationList = Arrays.asList(ApplyAggregation.LISTE);
		}
		final ProtocolAggregationData pad = new ProtocolAggregationData(attributes.size(), aggregationList);
		if(_archiveBuffer != null) // muss aber != null sein, sonst könnte es keine Aggregation geben...
		{
			pad.linkedAttributes = _archiveBuffer.getLinkedAttributes();
		}
		for(int i = 0; i < attributes.size(); i++) {
			final Elements element = attributes.get(i);

			pad.addAggregations(element.getAggregationClasses());

			if(element instanceof AttributeGroupDescription) {
				pad.columnDataType[i] = ((AttributeGroupDescription)element).qualifier.attributeGroup.getType().getId();
			}
			else if(element instanceof RealAttributeDescription) {
				pad.columnDataType[i] = ((RealAttributeDescription)element).getAttribute().getType().getId();
			}
			else if(element instanceof TempAttributeDescription) {
				pad.columnDataType[i] = ((TempAttributeDescription)element).getType().getId();
			}
			else {
				pad.columnDataType[i] = 0;
			}
		}
		if(_send) tryToSend(ProtocolAnswer.sendProtocol(_sender, _protocolId, _serializer, pad));
		saveProtocolData(pad);
	}

	/**
	 * Stellt den Administrationsdatensatz zusammen und sendet ihn an den Auftraggeber. Erzeugt und belegt {@link #_dataIdentifications}. Diese Methode darf erst
	 * NACH Starten des Archiv-/Onlinebuffers aufgerufen werden, da sonst {@link ProcessingBuffer#getTimeStampOrigin(TempAttributeDescription)} nicht verwendet
	 * werden kann.
	 *
	 * @throws FailureException Fehler bei der Kommunikation.
	 */
	private void sendAdministrationInformation() throws FailureException {
		_debug.finest("Sende Administrationsdatensatz für " + _protocolId);
		/*
		 * Zuordnung Spalten -> DI berechnen
		 */
		// [Spalte] -> Index der zugrundeliegenden Datenidentifikation in dids (weiter unten)
		final int[] columnToDataId = new int[_pi.getAttributes().size()];
		// Liste aller Dids in der Reihenfolge ihres Vorkommens in pi.getAttributes() -> Doppelte Einträge möglich
		final ArrayList<Qualifier> di = new ArrayList<Qualifier>();
		// Jeder Datenindex erhält eine eindeutige Nummer (= Index in dids)
		final ArrayList<Integer> map = new ArrayList<Integer>();
		// falls Archivbuffer fertig ist (z.B: weil leeres Intervall)...
		final ProcessingBuffer buf = (!_archiveBuffer.isDone()) ? _archiveBuffer : _onlineBuffer;

		// OPT
		int i = 0;
		for(final Elements element : _pi.getAttributes()) {
			if(element instanceof RealAttributeDescription) {
				final RealAttributeDescription rad = (RealAttributeDescription)element;
				if((columnToDataId[i] = di.indexOf(rad.qualifier)) == -1) {
					map.add(i);
					di.add(rad.qualifier);
					columnToDataId[i] = map.size() - 1;
				}
			}
			else if(element instanceof AttributeGroupDescription) {
				final AttributeGroupDescription ratgd = (AttributeGroupDescription)element;
				if((columnToDataId[i] = di.indexOf(ratgd.qualifier)) == -1) {
					map.add(i);
					di.add(ratgd.qualifier);
					columnToDataId[i] = map.size() - 1;
				}
			}
			else if(element instanceof TempAttributeDescription) {
				final TempAttributeDescription tad = (TempAttributeDescription)element;
				final RealAttributeDescription rad = buf.getTimeStampOrigin(tad);
				if(rad != null) {
					if((columnToDataId[i] = di.indexOf(rad.qualifier)) == -1) {
						map.add(i);
						di.add(rad.qualifier);
						columnToDataId[i] = map.size() - 1;
					}
				}
				else {
					columnToDataId[i] = ProtocolAdministrationData.NO_DID;
				}
			}
			else if(element instanceof DurationAttributeDescription) {
				columnToDataId[i] = ProtocolAdministrationData.NO_DID;
			}
			i++;
		}
		// map.toArray(dataIdentifications); geht nicht, da dataIdentifications sonst Integer[] sein müsste
		_dataIdentifications = new int[map.size()];
		for(int c = 0; c < _dataIdentifications.length; c++) {
			_dataIdentifications[c] = map.get(c);
		}

		final long[][] dids = new long[di.size()][3];
		for(int c = 0; c < di.size(); c++) {
			final Qualifier qualifier = di.get(c);
			if(qualifier.object != null){
				dids[c][0] = qualifier.object.getId();
			}
			dids[c][1] = qualifier.attributeGroup.getId();
			dids[c][2] = (qualifier.aspect != null) ? qualifier.aspect.getId() : -1; // aspect ist immer != null
		}

		final ProtocolAdministrationData pad = new ProtocolAdministrationData(
				_atlMeta.atlBase.name,
				_atlMeta.atlBase.date,
				_atlMeta.atlDetail.version,
				System.currentTimeMillis(),
				_configAuth.getName(),
				_pi.getAttributes(),
				columnToDataId,
				dids,
				_pi.getSimVar(),
				_configAuth.getPid(), _pp.getCreatorName(), _pp.getCreationDate(), _pp.getInfoText()
		);
		tryToSend(ProtocolAnswer.sendProtocol(_sender, _protocolId, _serializer, pad));
		saveProtocolData(pad);
	}

	/**
	 * Beginnt mit dem Sammeln von Daten vom Ars.
	 *
	 * @param stopCollectingData Zeitpunkt, bis zu dem Daten vom Archivsystem gesammelt werden.
	 *
	 * @throws InterruptedException Warten auf Archivantwort wurde unterbrochen.
	 */
	private void collectDataFromArs(final long stopCollectingData) throws FailureException, InterruptedException {
		final List<Tuple<Long, Long>> arsPeriods = getArsPeriods(stopCollectingData);

		_archiveBuffer = new ArchiveBuffer(_sender.getDav(), this, _configAuth, _pi, arsPeriods, getScriptName());
		_archiveBuffer.init();
	}

	/**
	 * Beginnt mit dem Sammeln von Daten vom Dav. Verwendet dazu den Online-Buffer.
	 *
	 * @return Zeitpunkt, bis zu dem die Daten aus dem Archiv geholt werden müssen.
	 *
	 * @throws FailureException Fehler.
	 */
	private long collectDataFromDav() throws FailureException {
		final long stopCollectingData;

		final long now = System.currentTimeMillis();

		final List<Tuple<Long, Long>> periods = getOnlinePeriods(now);

		_onlineBuffer = new OnlineBuffer(_sender.getDav(), this, _pi, periods, getOnlinePeriodIndex(now), getScriptName());
		_onlineBuffer.init();

		stopCollectingData = _onlineBuffer.getStartTime();

		return stopCollectingData;
	}

	/**
	 * Gibt den Index zurück, bei dem die Online-Zeitbereiche starten. Beispiel: Vergangenheit-Vergangenheit; Vergangenheit-Zukunft; Zukunft-Zukunft würde 1
	 * zurückgeben, da im zweiten Element angefangen wird, auf zukünftige Onlinedaten zuzugreifen.
	 * @param now Aktueller Zeitpunkt. Wird als Parameter übergeben, damit mehrere Aufrufe von {@link System#currentTimeMillis()} nicht unterschiedliche Ergebnisse liefern
	 * @return Index in _pi#getPeriods
	 */
	private int getOnlinePeriodIndex(final long now) {

		int n = 0;
		for(final Tuple<Long, Long> period : _pi.getPeriods()) {
			if(period.last <= now) {
				n++;
			}
		}
		return n;
	}

	/**
	 * Liefert die Zeiträume, in denen die Daten direkt vom Datenverteiler gesammelt werden sollen.
	 *
	 * @return Liste mit Zeiträumen.
	 * @param now
	 */
	private List<Tuple<Long, Long>> getOnlinePeriods(final long now) {
		final List<Tuple<Long, Long>> periods = new ArrayList<Tuple<Long, Long>>();

		for(final Tuple<Long, Long> period : _pi.getPeriods()) {
			if(period.last > now) periods.add(period);
		}
		return periods;
	}

	/**
	 * Liefert die Zeiträume, in denen die Daten vom Archivsystem angefragt werden sollen.
	 *
	 * @param stop Zeitpunkt bis zu dem Daten vom Archivsystem gesammelt werden sollen
	 *
	 * @return Zeiträume für die Daten vom Archivsystem geholt werden sollen.
	 */
	private List<Tuple<Long, Long>> getArsPeriods(final long stop) {
		final List<Tuple<Long, Long>> arsPeriods = new ArrayList<Tuple<Long, Long>>();
		Tuple<Long, Long> period;
		for(int i = 0; i < _pi.getPeriods().size(); i++) {
			period = _pi.getPeriods().get(i);
			if(period.last < stop) {
				arsPeriods.add(period);
			}
			else if(period.first <= stop) {
				arsPeriods.add(new Tuple<Long, Long>(period.first, stop));
			}
			else {
				break; // Zeitbereiche sind aufsteigend sortiert -> abbruch schon jetzt möglich
			}
		}

		return arsPeriods;
	}

	/**
	 * Bricht die Datenaufbereitung ab.
	 *
	 * @see de.bsvrz.pua.prot.processing.ProcessingInterface#abort()
	 */
	public void abort() {
		_abort = true;
		_jobState = JobState.Aborted;
		synchronized(this) {
			notifyAll();
		}

		// Datenquellen abbrechen

		if(_archiveBuffer != null) _archiveBuffer.abort();
		if(_onlineBuffer != null) _onlineBuffer.abort();

		// NotifyAll reicht nicht immer, da während Archivanfragen auf andere Objekte synchronisiert wird.
		// _archiveBuffer.abort() und _onlineBuffer.abort() rufen interrupt() auf ihre eigenen Threads auf, aber
		// nicht auf den ThreadElement-Thread, der ebenfalls hängen bleiben kann (in ArchiveUser.execute())
		_thread.interrupt();
	}

	/** @see de.bsvrz.pua.prot.processing.ProcessingInterface#getStatus() */
	public byte getStatus() {
		return _status;
	}


	/** @see de.bsvrz.pua.prot.processing.ProcessingInterface#getProtocolId() */
	public long getProtocolId() {
		return _protocolId;
	}

	@Override
	protected Object getClient() {
		return _sender.getObject();
	}

	@Override
	public long getJobId() {
		return _protocolId;
	}

	@Override
	public JobState getJobState() {
		return _jobState;
	}

	public void stopProtocolTransmission() {
		_send = false;
		String errorMsg = "Client hat die Übertragung abgebrochen";
		_debug.warning("Protokollerstellung wird abgebrochen :" + _protocolId + ", Grund: " + errorMsg);
		if(_send) try {
			tryToSend(ProtocolAnswer.sendAbort(_sender, _protocolId, errorMsg));
		}
		catch(FailureException e) {
			e.log();
		}
	}

	/**
	 * Liefert die Bezeichnung des Skripts zur Verwendung in Fehlerausgaben.
	 * @return Bezeichnung des Skripts
	 */
	private String getScriptName() {
		return _pp.getScript().toString();
	}

}
