/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.processing.onlinebuffer;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.util.ReferenceQualifier;
import de.bsvrz.pua.prot.util.attributes.AttributeIdentifier;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;

/**
 * Klasse zum Abfragen von Online-Daten.<br> Objekte dieser Klasse dürfen nicht zum synchronisieren verwendet werden. D.h. <code>notify</code> und
 * <code>notifyAll()</code> düfen nicht verwendet werden! (Diese Methoden können nicht überschrieben werden, da sie als <code>final</code> deklariert wurden.)
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class OnlineDataReceiver {

	/** Objekt das die Datensammlung durchführt. */
	private OnlineDataReceiverObject _receiver;

	protected OnlineDataReceiver(OnlineDataReceiverObject receiver) {
		this._receiver = receiver;
	}

	/**
	 * Liefert einen Datensatz. Die Methode wartet bis ein Datensatz eingetroffen ist.
	 *
	 * @param timeout Maximale Wartezeit. Wird die Wartezeit überschritten, so wird die Methode abgebrochen und <code>false</code> zurückgegeben. Ist dieser
	 *                Parameter <code>0</code> so wird solange gewartet, bis Daten eintreffen.
	 *
	 * @return Empfangener Datensatz.
	 *
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	public ExtResultData take(long timeout) throws InterruptedException {
		if(_receiver != null) {
			return _receiver.take(timeout, this);
		}
		else {
			return null;
		}
	}

	/** Meldet den Empfang von Daten ab. Das Objekt meldet sich vom Empfang der Daten ab. */
	public void unsubscribe() {
		if(_receiver != null) {
			_receiver.unsubscribe(this);
		}
		else {
			_receiver = null;
		}
	}
}


/**
 * Objekt zum Empfang von Online-Daten für Referenzattribute. Gehen Daten ein, so wird sofort die Referenzkette durchlaufen, um die endgültigen Daten zu finden
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 */
class OnlineDataReferenceReceiverObject extends OnlineDataReceiverObject {

	/** Qualifizierung der Referenz */
	private ReferenceQualifier qualifier;

	/** Verbindung zum Datenverteiler */
	private ClientDavInterface dav;

	/** Zeigt an auf welche Daten angemeldet werden soll. */
	private ReceiveOptions receiveOptions;

	/** Thread der die Referenzdaten einholt */
	private GetRefThread collectorThread;

	/**
	 * @param dav              Verbindung zum Datenverteiler
	 * @param receiveOptions   Empfangsoptionen
	 * @param qualifier        Datenidentifikation auf die sich der Receiver anmelden soll.
	 * @param receiverSet      Empfängergruppe, zu der dieser Empfänger gehört.
	 * @param newDataCollector Objekt das benachrichtig wird, wenn neue Daten vorliegen
	 */
	public OnlineDataReferenceReceiverObject(
			ClientDavInterface dav,
			ReceiveOptions receiveOptions,
			OnlineDataReceiverSet receiverSet,
			ReferenceQualifier qualifier,
			NewDataCollector newDataCollector) {
		super(receiverSet, qualifier, newDataCollector);
		this.dav = dav;
		this.receiveOptions = receiveOptions;
		this.qualifier = qualifier;
		collectorThread = new GetRefThread(this);
		collectorThread.start();
	}

	/**
	 * Speichert den eingegangenen Datensatz. Da es sich um einen Empfänger für Referenzdaten handelt, wird die Referenzkette durchlaufen, und es wird der Wert der
	 * Referenz gespeichert. Da diese Methode vom der Dav-Callback-Methode {@link OnlineDataReceiverObject#update(ResultData[])} aufgerufen wird, muss ein neuer
	 * Thread gestartet werden, wenn eine Referenz abgeholt wird - sonst meldet der Dav keine weiteren Daten mehr.
	 *
	 * @see de.bsvrz.pua.prot.processing.onlinebuffer.OnlineDataReceiverObject#storeData(ResultData)
	 */
	protected void getData(final ResultData result) {
		collectorThread.add(result);
	}

	/**
	 * Holt die Referenz vom Datenverteiler
	 *
	 * @param reference Datenidentifikation, die benötigt wird. (Das Objekt reference.object) wird nicht verwendet!
	 * @param object    Objekt, für das die Daten abgerufen werden sollen.
	 * @param dataTime  Zeitstempel des ursprünglichen Datensatzes
	 *
	 * @return Wert der Referenz
	 *
	 * @throws FailureException Fehler beim Datenempfang
	 */
	private ResultData takeReference(ReferenceQualifier reference, SystemObject object, long dataTime) throws FailureException {
		ResultData value = null;
		ReferenceReceiver receiver = null;
		try {
			receiver = new ReferenceReceiver(dav, object, reference.attributeGroup, reference.aspect, receiveOptions);
			try {
				value = receiver.take(500);
			}
			catch(InterruptedException e) {
				;
			}
			try {
				if(receiver != null) receiver.unsubscribe();
			}
			catch(FailureException e) {
				e.log();
			}

			if(value != null) // falls value == null -> finally block
			{
				if(value.getDataTime() > dataTime) {
					value = null;
				}
				else if(reference.getReference() != null && value.hasData()) // nur weitermachen, wenn es auch Daten
				// gibt...
				{
					return takeReference(reference.getReference(), getReference(reference.getAttributeIdent(), value), dataTime);
				}
				else {
					return value;
				}
			}
		}
		catch(FailureException e) {
			e.log();
		}
		finally {
			if(value == null) {
				// Ergebnisdatensatz selbst zusammenbauen, aber ohne Daten
				value = new ResultData(object, new DataDescription(qualifier.attributeGroup, qualifier.aspect), dataTime, null);
			}
		}
		return value;
	}

	/** @param object  */
	@Override
    public void unsubscribe(Object object) {
		collectorThread.abort();
		super.unsubscribe(object);
	}

	/**
	 * Holt eine Refernz aus einem Datensatz.
	 *
	 * @param attributeIdentifier       Attribut (bzw. Attributkette) hinter der der Referenzwert liegt
	 * @param resultData Daten.
	 *
	 * @return Referenziertes Objekt.
	 *
	 * @throws FailureException Fehler beim auslesen der Referenz.
	 */
	private static SystemObject getReference(AttributeIdentifier attributeIdentifier, ResultData resultData) throws FailureException {
		try {
			return attributeIdentifier.getDataElement(resultData.getData()).asReferenceValue().getSystemObject();
		}
		catch(NullPointerException e) {
			throw new FailureException(ErrorMessage.UNEXPECTED_DATA_TYPE + resultData.getData() + ", " + attributeIdentifier, e, LoggerException.WARNING);
		}
	}

	class GetRefThread extends Thread {

		private final OnlineDataReferenceReceiverObject object;

		private RingBuffer<ResultData> incomingResultData = new RingBuffer<ResultData>(100, RingBuffer.UNBOUNDED_SIZE);

		boolean abort = false;

		/** @param object  */
		public GetRefThread(OnlineDataReferenceReceiverObject object) {
			this.object = object;
		}

		/** Beendet den Thread */
		public void abort() {
			abort = true;
			this.interrupt();
		}

		/** @see java.lang.Runnable#run() */
		@Override
        public void run() {
			ResultData result;
			while(!this.isInterrupted() && !abort) {
				try {
					result = incomingResultData.pop();
					SystemObject so = getReference(qualifier.getAttributeIdent(), result);
					ResultData resultData = takeReference(qualifier.getReference(), so, result.getDataTime());
					object.storeData(resultData);
				}
				catch(FailureException e) {
					e.log();
				}
				catch(InterruptedException e) {
				}
			}
		}

		/**
		 * Weiteren Datensatz hinzufügen
		 *
		 * @param data
		 */
		public void add(ResultData data) {
			incomingResultData.push(data);
		}
	}
}

/**
 * Klasse dient zum Abfragen von aktuellen Datenwerten auf dem Dav. Wird für Referenzen verwendet. {@link #unsubscribe()} muss aufgerufen werden, nachdem die
 * Daten mit {@link #take(long)} abgerufen wurden.
 *
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
class ReferenceReceiver implements ClientReceiverInterface {

	/** Zeigt an ob Daten vorliegen. */
	private boolean receivedData = false;

	/** Daten. */
	private ResultData resultData = null;

	/** Verbindung zum Datenverteiler */
	private ClientDavInterface dav;

	private SystemObject object;

	private DataDescription dataDescription;

	/**
	 * Meldet den Empfang von Daten an.
	 *
	 * @param dav            Verbindung zum Datenverteiler.
	 * @param systemObject   Objekt.
	 * @param attributeGroup Attributgruppe.
	 * @param aspect         Aspekt.
	 * @param options        Datenart.
	 *
	 * @throws FailureException Fehler beim Anmelden auf die Datenidentifikation.
	 */
    public ReferenceReceiver(ClientDavInterface dav, SystemObject systemObject, AttributeGroup attributeGroup, Aspect aspect, ReceiveOptions options)
			throws FailureException {
		this.dav = dav;
		this.object = systemObject;
		this.dataDescription = new DataDescription(attributeGroup, aspect);
		ConnectionManager.subscribeReceiver(dav, this, systemObject, dataDescription, options, ReceiverRole.receiver());
	}

	/**
	 * Wird verwendet um Daten von Referenzen vom Dav abzuholen.
	 *
	 * @see ClientReceiverInterface#update(ResultData[])
	 */
	public void update(ResultData[] results) {
		if(!receivedData) {
			resultData = results[results.length - 1];
			receivedData = true;
			synchronized(this) {
				notifyAll();
			}
		}
	}

	/**
	 * Meldet den Empfang von Daten ab.
	 *
	 * @throws FailureException Fehler bei der Kommunikation mit der Konfiguration
	 */
    public void unsubscribe() throws FailureException {
		ConnectionManager.unsubscribeReceiver(dav, this, object, dataDescription);
	}

	/**
	 * Gibt den aktuellen Wert der Datenidentifikation zurück. Meldet sich anschließend vom Datenverteiler ab.
	 *
	 * @param timeout Maximale Wartezeit.
	 *
	 * @return Aktueller Wert oder <code>null</code> falls keine Daten innerhalb der Wartezeit vorliegen.
	 *
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	public ResultData take(long timeout) throws InterruptedException {
		if(!receivedData) // keine while-Schleife, sonst wird der Timeout nicht berücksichtigt
		{
			synchronized(this) {
				wait(timeout);
			}
		}
		// TODO hier Zeitstempel der Referenz überprüfen!
		if(receivedData) {
			return resultData;
		}
		else {
			return null;
		}
	}

	/** @return Zeigt an ob Daten vorliegen. */
	public boolean hasData() {
		return receivedData;
	}
}
