/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.util;

import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.filters.PostFilter;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.manager.ManagerErrorMessage;
import de.bsvrz.pua.prot.util.attributes.*;
import de.bsvrz.sys.funclib.losb.datk.NoChangeMarker;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;
import de.bsvrz.sys.funclib.losb.util.Tuple;

import java.util.*;

/**
 * Diese Klasse modelliert die aus dem Skript extrahierten Informationen, die zur Datenaufbereitung nötig sind. ACHTUNG: Änderungen an dieser Klasse müssen im
 * Feinentwurf berücksichtigt werden!
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ProcessingInformation {

	/**
	 * Die Pseudo-Objekt-zuordnungen
	 */
	private Map<String, String[]> _pseudoObjects;

	/**
	 * Die gebundenen Aspekte
	 */
	private Map<String, String> _aspectBindings;

	/**
	 * Alle Aspekte die anbindbar sind, unabhängig davon, ob diese gebunden sind.
	 */
	private List<String> _aspects;

	/**
	 * Die im Skript definierten Aliase
	 */
	private HashMap<String,String> _aliases;

	public void setPseudoObjects(final Map<String, String[]> pseudoObjects) {
		_pseudoObjects = pseudoObjects;
	}

	public Map<String, String[]> getPseudoObjects() {
		if(_pseudoObjects == null) return Collections.emptyMap();
		return Collections.unmodifiableMap(_pseudoObjects);
	}

	public void setAspectBindings(final Map<String, String> aspectBindings) {
		_aspectBindings = aspectBindings;
	}

	public Map<String, String> getAspectBindings() {
		if(_aspectBindings == null) return Collections.emptyMap();
		return Collections.unmodifiableMap(_aspectBindings);
	}

	public void setPossibleAspectBindings(final List<String> possibleAspectBindings) {
		_aspects = possibleAspectBindings;
	}

	public List<String> getPossibleAspectBindings() {
		if(_aspects == null) return Collections.emptyList();
		return Collections.unmodifiableList(_aspects);
	}

	public void setAliases(final HashMap<String, String> aliases) {
		_aliases = aliases;
	}

	public Map<String, String> getAliases() {
		if(_aliases == null) return Collections.emptyMap();
		return Collections.unmodifiableMap(_aliases);
	}

	/**
	 * @param deltaProtocol DeltaProtocol
	 * @deprecated Bitte {@link #setProtocolType(de.bsvrz.sys.funclib.losb.datk.ProtocolType)} benutzen
	 */
	@Deprecated
	public void setDeltaProtocol(boolean deltaProtocol) {
		setProtocolType(deltaProtocol ? ProtocolType.DeltaProtocol : ProtocolType.StatusProtocol);
	}


	/**
	 * Gibt <tt>true</tt> zurück, wenn es sich um ein Änderungsprotokoll handelt
	 * @return <tt>true</tt>, wenn es sich um ein Änderungsprotokoll handelt, sonst <tt>false</tt>
	 * @deprecated Bitte {@link #getProtocolType()} benutzen
	 */
	@Deprecated
	public boolean isDeltaProtocol() {
		return getProtocolType() == ProtocolType.DeltaProtocol;
	}

	@Deprecated
	public Map<SystemObjectType, SystemObject> getObjects() {
		if(_mainObject == null) return Collections.emptyMap();
		return Collections.singletonMap(_mainObject.getType(), _mainObject);
	}

	/** Mögliche Anwendungsarten für die Aggregationen. */
	public enum ApplyAggregation { // Die Reihenfolge NICHT verändern, da diese in setOrder() verwendet wird, um zu sortieren.
		/** Gesamt (Objekt) Aggregation */
		GESAMT_OBJEKT,
		/** Gesamt (Spalte) Aggregation */
		GESAMT_SPALTE,
		/** Intervall (Objekt) Aggregation */
		INTERVALL_OBJEKT,
		/** Intervall (Spalte) Aggregation */
		INTERVALL_SPALTE,
		/** Liste Aggregation */
		LISTE
	}

	/** Liste der Attribute (reale Attribute, temporäre Attribute) in der Reihenfolge ihres Auftretens im Spaltenbereich des Skripts. */
	private List<Elements> _elements;

	/** Filter, bestehend aus Filter - Ausdruck und Boolean Wert, der angibt, ob verschmolzen werden soll */
	private Tuple<ExpressionTree, Boolean> _constraint;

	/** Benutzerdefinierter Filter */
	private PostFilter _postFilter;

	/** Liste aller Datenidentifikationen, die beim Archivsystem angefragt werden müssen. */
	private HashSet<RealElement> _allRealElements;

	/** Legt fest wie die Aggregationen angewendet werden. (Spaltenweise, Zeitbereichs-Übergreifend, etc) Sortierung ist von grob nach fein. */
	private List<ApplyAggregation> _order;

	/** Zeitbereiche für die das Protokoll erstellt werden soll */
	private List<Tuple<Long, Long>> _periods;

	/** Simulationsvariante für das Protokoll */
	private short _simVar;

	/** True falls das Protokoll gespeichert werden soll. Sonst false */
	private boolean _saveProtocol;

	/**
	 * Typ des Protokolls. Nach dem Parsen des Skriptes ist hier in jedem Fall ein Wert außer {@link de.bsvrz.sys.funclib.losb.datk.ProtocolType#Undefined) gespeichert (siehe
	 * {@link de.bsvrz.pua.prot.interpreter.semantics.Semantics#cleanUp(ProcessingParameter, boolean)}). Der null-Wert ist hier nicht erlaubt.
	 */
	private ProtocolType _protocolType = ProtocolType.Undefined;

	/**
	 * markierungsart für "Keine Ädnerungen"-Datensätze. Nach dem Parsen des Skriptes ist hier in jedem Fall ein Wert außer {@link de.bsvrz.sys.funclib.losb.datk.NoChangeMarker#Undefined) gespeichert (siehe
	 * {@link de.bsvrz.pua.prot.interpreter.semantics.Semantics#cleanUp(ProcessingParameter, boolean)}). Der null-Wert ist hier nicht erlaubt.
	 */
	private NoChangeMarker _noChangeMarker = NoChangeMarker.Undefined;

	/** Art der Daten, die vom Archivsystem angefordert werden sollen */
	private ArchiveDataKindCombination _archiveDataKinds = null;

	/** Hauptobjekt des Protokolls. */
	private SystemObject _mainObject;

	/**
	 * Enthält alle Elemente, die direkt oder indirekt im Spaltenbereich vorkommen. Inidrekt kommen z.B. die von temporären referenzierten realen Attribute im
	 * Spaltenbereich vor. Alle Elemente, die nicht in dieser Map sind, sind auch nicht im Spaltenbereich.
	 */
	private HashSet<Object> _shownElements = null;

	/** Erzeugt das Objekt. */
	public ProcessingInformation() {
		_elements = new ArrayList<Elements>();
		_constraint = null;
		_postFilter = null;
		_order = new ArrayList<ApplyAggregation>();
		_periods = null;
		_simVar = Qualifier.INVALID_SIMVAR;
		_saveProtocol = false;
		_archiveDataKinds = null;
		_mainObject = null;
	}

	/**
	 * Erzeugt das Objekt.
	 *
	 * @param mainObject Hauptobjekt des Protokolls.
	 */
	public ProcessingInformation(final SystemObject mainObject) {
		this();
		this._mainObject = mainObject;
	}

	/** @return Liefert alle Attribute, die im "Spalten" Bereich definiert wurden. */
	public List<Elements> getAttributes() {
		return Collections.unmodifiableList(_elements);
	}

	/** @param attributes Attribute die im "Spalten" Bereich definiert wurden. */
	public void setAttributes(final List<Elements> attributes) {
		this._elements = attributes;
	}

	/**
	 * Fügt ein weiteres Elements hinzu. ACHTUNG: Die Liste der Datenindizes wird NICHT neu berechnet. Daher muss ggf. {@link #update()} aufgerufen werden.
	 *
	 * @param attribute Attribute
	 */
	public void addElement(final Elements attribute) {
		_elements.add(attribute);
	}

	public void replaceElement(final int index, final List<? extends TempAttributeDescription> elementsList, final String columnName) {
		if(elementsList.size() == 1) {
			replaceElement(index, new TempAttributeDescription(elementsList.get(0), columnName));
		}
		else {
			_elements.set(index, elementsList.get(0));
			_elements.addAll(index + 1, elementsList.subList(1, elementsList.size()));
			for(int i = index; i < index + elementsList.size(); i++) {
				_elements.set(i, new TempAttributeDescription((TempAttributeDescription)_elements.get(i), columnName));
			}
		}
	}

	public void replaceElement(final int index, final TempAttributeDescription element) {
		_elements.set(index, element);
	}

	/**
	 * Neuberechnung der Liste der Elemente, die vom Archivsystem angefragt werden sollen. Diese Elemente können mittels {@link #getAllRealElements()} abgerufen
	 * werden. Diese Liste enthält alle realen Attribute (aus dem Spaltenbereich und aus Ausdrücken (temporäre Attribute, Constraint) und Attributgruppen.
	 * Berechnet außerdem, welche Elemente im Spaltenbereich vorkommen. Diese Information kann dann mit {@link #isShown(de.bsvrz.pua.prot.util.attributes.RealElement)}
	 * )} abgerufen werden. <b>Wichtig</b>: Diese Methode darf erst aufgerufen werden, wenn sichergestellt ist, dass KEINE REKURSIONEN in den Ausdrücken verwendet
	 * werden! D.h. es <b>muss vorher</b> {@link de.bsvrz.pua.prot.interpreter.semantics.TempAttributeSemantics#cleanUp(ProcessingParameter, boolean)} aufgerufen
	 * werden!
	 */
	@SuppressWarnings("unchecked")
	public void update() {
		_allRealElements = new HashSet<RealElement>(_elements.size());
		_shownElements = new HashSet<Object>((int)(_elements.size() * 1.3));
		// Spalten-Definition.
		for(final Elements att : _elements) {
			final Set<? extends RealElement> realElements = att.getRealElements();
			_allRealElements.addAll(realElements);
			_shownElements.addAll(realElements);
			_shownElements.add(att);
		}

		// zusätzlich die benötigten Attribute aus dem Filterausdruck holen
		if(_constraint != null) {
			final List<AttributeDescription> exprAttributes = _constraint.first.getAttributes();
			for(final AttributeDescription att : exprAttributes) {
				_allRealElements.addAll(att.getRealElements());
			}
		}
	}

	/** @return Einschränkung. */
	public Tuple<ExpressionTree, Boolean> getConstraint() {
		return _constraint;
	}

	/**
	 * Erzeugt zusätzlich Liste mit Datenindices, die vom Archivsystem angefragt werden sollen und über getDataIndices() abgefragt werden können.
	 *
	 * @param constraint Einschränkung
	 */
	public void setConstraint(final Tuple<ExpressionTree, Boolean> constraint) {
		this._constraint = constraint;
	}

	/**
	 * Berechnet die für das Skript benötigten Qualifizierungen.
	 *
	 * @return Alle im Skript verwendeten Datenidentifikationen.
	 */
	public HashSet<Qualifier> getDataIndices() {
		final HashSet<Qualifier> dataIndices = new HashSet<Qualifier>();
		for(final Elements att : _elements)
		// DI aus Attribut(gruppe) extrahieren
		{
			dataIndices.addAll(att.getQualifiers());
		}

		// zusätzlich die benötigten Attribute aus dem Filterausdruck holen
		if(_constraint != null) {
			final List<AttributeDescription> exprAttributes = _constraint.first.getAttributes();
			for(final AttributeDescription att : exprAttributes) {
				dataIndices.addAll(att.getQualifiers());
			}
		}
		return dataIndices;
	}

	/** @return Anwendung der Aggregation. */
	public List<ApplyAggregation> getOrder() {
		return _order;
	}

	/**
	 * Ordnet den Vektor der Aggregations-Anwendungsvorschriften von grob nach fein
	 *
	 * @param rules Anwendungsvorschrift für die Aggregationen
	 *
	 * @throws SemanticErrorException Falls Wiederholungen in der anwendung auftreten (etwa weil gleichzeitig sowohl gesamt(objekt) als auch gesamt(spalte)
	 *                                verwendet wurde)
	 */
	public void setOrder(final HashSet<ApplyAggregation> rules) throws SemanticErrorException {
		this._order = new ArrayList<ApplyAggregation>();

		if(rules.contains(ApplyAggregation.GESAMT_OBJEKT)) _order.add(ApplyAggregation.GESAMT_OBJEKT);
		if(rules.contains(ApplyAggregation.GESAMT_SPALTE)) _order.add(ApplyAggregation.GESAMT_SPALTE);

		if(rules.contains(ApplyAggregation.INTERVALL_OBJEKT)) _order.add(ApplyAggregation.INTERVALL_OBJEKT);
		if(rules.contains(ApplyAggregation.INTERVALL_SPALTE)) _order.add(ApplyAggregation.INTERVALL_SPALTE);

		if(rules.contains(ApplyAggregation.LISTE)) _order.add(ApplyAggregation.LISTE);
	}

	/** @return Nachfilter. */
	public PostFilter getUserFilter() {
		return _postFilter;
	}

	/** @param userFilter Nachfilter. */
	public void setUserFilter(final PostFilter userFilter) {
		this._postFilter = userFilter;
	}

	/**
	 * Legt den Nachfilter fest.
	 *
	 * @param filter Klassenname des Nachfilters
	 *
	 * @return Null oder eine Fehlermeldung im Fehlerfall.
	 */
	public String setPostFilter(final String filter) {
		String error = null;
		Class<?> c = null;
		Object o = null;

		try {
			c = Class.forName(filter);
			o = c.newInstance();
			if(o instanceof PostFilter) {
				this._postFilter = (PostFilter)o;
			}
			else {
				error = InterpreterErrorMessage.WRONG_CLASSTYPE + c.getName() + ", " + this._postFilter.getClass().getName();
			}
		}
		catch(ClassNotFoundException e) {
			error = InterpreterErrorMessage.WRONG_CLASSNAME + filter;
		}
		catch(Exception e) {
			error = InterpreterErrorMessage.CAN_NOT_CREATE_CLASS + filter + ", " + e.getMessage();
		}

		return error;
	}

	/** @return Zeitbereiche für die das Protokoll erstellt wwird. */
	public List<Tuple<Long, Long>> getPeriods() {
		if(_periods == null) return Collections.emptyList();
		return _periods;
	}

	/** @param periods Zeitbereiche für die das Protokoll erstellt wwird. */
	public void setPeriods(final List<Tuple<Long, Long>> periods) {
		this._periods = periods;
	}

	/** @return Simulationsvariante. */
	public short getSimVar() {
		return _simVar;
	}

	/**
	 * Legt Simulationsvariante fest. Es wird nicht geprüft, ob die Simulationsvariante gültig ist.
	 *
	 * @param simVar Simulationsvariante
	 */
	public void setSimVar(final short simVar) {
		this._simVar = simVar;
	}


	/**
	 * Gibt den Protokolltyp zurück. Falls das Script noch nicht verarbeitet wurde kann der Wert {@link de.bsvrz.sys.funclib.losb.datk.ProtocolType#Undefined}
	 * sein, sonst ist auf jeden Fall ein gültiger ProtocolType gespeichert.
	 * @return ProtocolType
	 */
	public ProtocolType getProtocolType() {
		return _protocolType;
	}

	/**
	 * Setzt die Protokollart
	 * @param protocolType neue Protokollart (!= null)
	 */
	public void setProtocolType(final ProtocolType protocolType) {
		if(protocolType == null) throw new IllegalArgumentException("protocolType ist null");
		this._protocolType = protocolType;
	}

	/**
	 * Setzt die Übertragungsart von "Keine Änderungen"-Datensätzen
	 * @param noChangeMarker Übertragungsart von "Keine Änderungen"-Datensätzen
	 */
	public void setNoChangeMarker(final NoChangeMarker noChangeMarker) {
		this._noChangeMarker = noChangeMarker;
	}

	/**
	 * Gibt die Übertragungsart von "Keine Änderungen"-Datensätzen zurück.
	 * Falls das Script noch nicht verarbeitet wurde kann der Wert {@link de.bsvrz.sys.funclib.losb.datk.NoChangeMarker#Undefined}
	 * sein, sonst ist auf jeden Fall ein gültiger NoChangeMarker gespeichert.
	 * @return NoChangeMarker
	 */
	public NoChangeMarker getNoChangeMarker() {
		return _noChangeMarker;
	}

	/** @return Gibt zurück ob das Protokoll gespeichert werden soll */
	public boolean isSaveProtocol() {
		return _saveProtocol;
	}

	/** @param saveProtocol Setzt, ob das Protokoll gespeichert werden soll */
	public void setSaveProtocol(final boolean saveProtocol) {
		this._saveProtocol = saveProtocol;
	}

	/**
	 * HTML Ausgabe.
	 *
	 * @param name Name.
	 *
	 * @return String mit HTML Ausgabe.
	 */
	public String toHtml(final String name) {
		final StringBuffer sb = new StringBuffer();

		sb.append(
				"<head><title>ProcessingInformation fuer " + name + "</title></head><body><di><p align=center ><span><b>ProcessingInformation für " + name
				+ "</b></span><b></b></p><p align=center><b></b></p>"
		);

		sb.append("<b>Elemente:</b><ul>");
		for(final Elements e : _elements) {
			sb.append("<li>" + e + "</li>");
		}
		sb.append("</ul>");

		sb.append("<b>Constraint:</b>");
		sb.append("<blockquote>" + _constraint + "</blockquote>");
		final HashSet<Qualifier> dataIndices = getDataIndices();
		if(dataIndices != null && !dataIndices.isEmpty()) {
			sb.append("<b>DataIndices:</b><ul>");
			final Iterator<Qualifier> i = dataIndices.iterator();
			while(i.hasNext()) sb.append("<li>" + i.next() + "</li>");
			sb.append("</ul>");
		}
		else {
			sb.append("<b>Keine Informationen über dataIndices</b><p></p>");
		}

		sb.append("<b>Deltaprotocol:</b>");
		sb.append("<blockquote>" + _protocolType + "</blockquote>");

		sb.append("<b>Order:</b>");
		sb.append("<blockquote>" + _order + "</blockquote>");

		if(_periods != null) {
			sb.append("<b>periods:</b><ul>");
			for(final Tuple<Long, Long> p : _periods) {
				sb.append("<li>" + p + "</li>");
			}
			sb.append("</ul>");
		}
		else {
			sb.append("<b>Keine Informationen über periods</b><p></p>");
		}

		sb.append("<b>saveProtocol:</b>");
		sb.append("<blockquote>" + _saveProtocol + "</blockquote>");

		sb.append("<b>simVar:</b>");
		sb.append("<blockquote>" + _simVar + "</blockquote>");

		sb.append("<b>postFilter:</b>");
		sb.append("<blockquote>" + _postFilter + "</blockquote>");

		sb.append("</body></html>");

		return sb.toString();
	}

	/** @return Returns the archiveDataKind. */
	public ArchiveDataKindCombination getArchiveDataKinds() {
		return _archiveDataKinds;
	}

	/**
	 * Wirft eine Exception falls das Feld leer ist oder null.
	 *
	 * @param archiveDataKindCombination The archiveDataKindCombination to set.
	 *
	 * @throws FailureException Das Feld ist null oder leer
	 */
	public void setArchiveDataKinds(final ArchiveDataKindCombination archiveDataKindCombination) throws FailureException {
		if(archiveDataKindCombination == null) throw new FailureException(ErrorMessage.IS_NULL_OR_EMPTY, LoggerException.WARNING);
		this._archiveDataKinds = archiveDataKindCombination;
	}

	/**
	 * Bezieht alle Membervariablen des Objekts in den Vergleich mit ein.
	 *
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(final Object obj) {
		if(obj instanceof ProcessingInformation) {
			final ProcessingInformation pi = (ProcessingInformation)obj;
			final boolean result;

			if(this._postFilter != null && pi._postFilter != null) {
				result = this._postFilter.getClass().getName().equals(pi._postFilter.getClass().getName());
			}
			else if((this._postFilter != null && pi._postFilter == null) || (this._postFilter == null && pi._postFilter != null)) {
				return false;
			}
			else {
				result = true;
			}

			return result && this._protocolType == pi._protocolType && Objects.equals(this._elements, pi._elements) && Objects.equals(this._order, pi._order)
			       && Objects.equals(this._periods, pi._periods) && Objects.equals(this._constraint, pi._constraint)
			       && Objects.equals(this._archiveDataKinds, pi._archiveDataKinds) && this._saveProtocol == pi._saveProtocol;
		}
		else {
			return false;
		}
	}

	/**
	 * Liefert eine Liste der realen Attribute und Attributgruppen, die vom Archivsystem angefragt werden sollen.
	 *
	 * @return Liste mit allen realen Attributen aus dem Spaltenbereich und aus den temporären Attributen und Attributgruppen.
	 */
	public HashSet<RealElement> getAllRealElements() {
		if(_allRealElements == null) update();
		return _allRealElements;
	}

	/**
	 * Setzt das Hauptobjekt des Protokolls.
	 *
	 * @param mainObject Hauptobjekt des Protokolls.
	 */
	public void setMainObject(final SystemObject mainObject) {
		this._mainObject = mainObject;
	}

	/**
	 * Gibt das Hauptobjekt des Protokolls zurück.
	 *
	 * @return Hauptobjekt.
	 */
	public SystemObject getMainObject() {
		return _mainObject;
	}

	/**
	 * Zeigt an ob das Element im Spalten-Bereich verwendet wird.
	 *
	 * @param obj Element
	 *
	 * @return <code>true</code> falls das Element im Spalten-Bereich verwendet wird. <code>false</code> sonst.
	 */
	public boolean isShown(final RealElement obj) {
		return _shownElements.contains(obj);
	}

	/**
	 * Liefert die Positionen der Elemente, die keine Status-information bereitstellen und bei der Berechnung des Status dehalb ignoriert werden sollte
	 * (Zeitdauer-Attribute und Platzhalterattribute)
	 *
	 * @return Positionen der Zeitdauer- und Platzhalter-Attribute. Wird kein Zeitdauer-/Platzhalter-Attribut verwendet, dann wird ein leeres Set zurückgegeben.
	 */
	public Set<Integer> getPositionsToIgnoreForStatus() {
		final HashSet<Integer> positions = new HashSet<Integer>();
		for(int i = 0; i < _elements.size(); i++) {
			final Elements element = _elements.get(i);
			if(element instanceof DurationAttributeDescription) {
				positions.add(i);
			}
			else if(element instanceof RealAttributeDescription) {
				final RealAttributeDescription realAttributeDescription = (RealAttributeDescription)element;
				if(realAttributeDescription.qualifier.object == null) {
					positions.add(i);
				}
			}
		}
		return positions;
	}

	/**
	 * Liefert alle realen Attribute zurück, die im Skript verwendet werden.
	 *
	 * @return Relae Attribute die im Skript verwendet werden. Es werden nicht nur die Attribute aus dem Spalten-Bereich, sondern auch die Attribute aus dem
	 *         Definitionen-Bereich zurückgeliefert.
	 */
	public Set<RealAttributeDescription> getAllRealAttributes() {
		final HashSet<RealAttributeDescription> rads = new HashSet<RealAttributeDescription>();
		for(final RealElement re : getAllRealElements()) {
			if(re instanceof RealAttributeDescription) rads.add((RealAttributeDescription)re);
		}
		return rads;
	}
}
