/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.util.attributes;

import de.bsvrz.dav.daf.main.config.Attribute;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.ReferenceAttributeType;
import de.bsvrz.pua.prot.aggregations.Aggregation;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.ScopeException;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.processing.archivebuffer.RealArchiveAttribute;
import de.bsvrz.pua.prot.util.Qualifier;
import de.bsvrz.pua.prot.util.ReferenceQualifier;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;

import java.util.*;

/**
 * Beschreibt ein reales Attribut.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class RealAttributeDescription extends AttributeDescription implements RealElement {

	public Qualifier qualifier;

	private RealAttributeDescription _reference = null;

	private AttributeIdentifier _attributeIdentifier = null;

	/**
	 * <code>true</code> falls Daten durch die Datenaufbereitung aufgefüllt werden sollen. Wird in {@link #equals(Object)} und {@link #hashCode()} nicht
	 * berücksichtigt.
	 */
	private boolean _padData = true;

	private int _index = -1;

	private String _alias = null;

	/**
	 * @param simVar  Simulationsvariante.
	 * @param padData <code>true</code> falls Daten durch die Datenaufbereitung aufgefüllt werden sollen.
	 */
	public RealAttributeDescription(final short simVar, final boolean padData) {
		this.qualifier = new Qualifier(simVar);
		this._padData = padData;
	}

	/**
	 * Konstruktor zu Testzwecken. Daten werden aufgefüllt.
	 *
	 * @param model       Konfiguration.
	 * @param so          Systemobjekt.
	 * @param atg         Attributgruppe.
	 * @param asp         Aspekt.
	 * @param columnName  Spaltenname. Kann <code>null</code> sein.
	 * @param aggregation Zu verwendende Aggregation. Kann <code>null</code> sein.
	 * @param name        Name des Attributs. Kann <code>null</code> sein.
	 * @param simVar      Simulationsvariante.
	 *
	 * @throws FailureException Ungültige Qualifizierung.
	 */
	@Deprecated
	public RealAttributeDescription(final DataModel model, final String so, final String atg, final String asp, final String name, final short simVar, final String columnName, final List<Class<? extends Aggregation>> aggregation)
			throws FailureException {
		super(columnName, aggregation, name);
		this.qualifier = new Qualifier(model, so, atg, asp, simVar);
		try {
			this._attributeIdentifier = new AttributeIdentifier(name, model.getAttributeGroup(atg));
		}
		catch(SemanticErrorException e) {
			throw new IllegalArgumentException(e);
		}
	}

	/**
	 * Vergleicht zwei RealAttributeDescription Objekte. Dabei werden alle Member-Variablen (außer {@link Elements#_columnName} und {@link Elements#_aggregationClasses})
	 * verglichen Object darf auch vom Typ {@link de.bsvrz.pua.prot.processing.archivebuffer.RealArchiveAttribute} sein. In diesem Fall wird nur die zugeordnete
	 * Attributbeschreibung verglichen. Dies ist notwendig, damit diese beiden Typen austauschbar in einer Hashtable verwendet werden können.
	 *
	 * @param obj Vergleichsobjekt
	 *
	 * @return True falls die Objekte bis auf columnName und aggregation übereinstimmen, false sonst
	 */
	@Override
	public boolean equals(final Object obj) {
		boolean result = false;

		final RealAttributeDescription ad;
		if(obj instanceof RealAttributeDescription) {
			ad = (RealAttributeDescription)obj;
			result = Objects.equals(qualifier, ad.qualifier) && Objects.equals(getName(), ad.getName()) && Objects.equals(_reference, ad._reference)
			         && Objects.equals(this._attributeIdentifier, this._attributeIdentifier) && this._padData == ad._padData;
		}
		else if(obj instanceof RealArchiveAttribute) result = ((RealArchiveAttribute)obj).equals(this);
		return result;
	}

	/**
	 * Gibt den verwendeten Dataindex zurück
	 *
	 * @return Datenindex des AttributeDescription
	 */
	@Override
	public Collection<Qualifier> getQualifiers() {
		final HashSet<Qualifier> h = new HashSet<Qualifier>();
		h.add(qualifier);

		return h;
	}

	/**
	 * Hashwert des Objekts ist abhängig von allen Membervariablen. (außer columnName und aggregation)
	 *
	 * @return Hashwert des Objekts
	 */
	@Override
	public int hashCode() {
		int result = 0;

		if(qualifier != null) result = qualifier.hashCode();
		if(getName() != null) result = 37 * result + getName().hashCode();
		if(_reference != null) result = 37 * result + _reference.hashCode();
		if(this._attributeIdentifier != null) result = 37 * result + this._attributeIdentifier.hashCode();
		if(!_padData) result *= 37;

		return result;
	}

	/** @see java.lang.Object#toString() */
	@Override
	public String toString() {
		String result = "Reales " + ((_padData) ? "" : "nicht ") + "aufzufüllendes Attribut: ";

		if(qualifier != null) result = result + qualifier.toString() + "::" + getAttributeIdentifier();

		final RealAttributeDescription ref = _reference;
		if(ref != null) {
			result += " => " + ref.toString();
		}

		if(getColumnName() != null) result += " als \"" + getColumnName() + "\"";
		if(getAggregationClasses() != null) result += ", aggregiert durch " + getAggregationClasses();

		return result;
	}

	/**
	 * Klont die Attributbeschreibung.
	 *
	 * @return Geklonte Attributbeschreibung.
	 */
	@Override
	public RealAttributeDescription clone() {
		
		final RealAttributeDescription clone = new RealAttributeDescription(this.qualifier.simVar, this._padData);

		if(qualifier != null) clone.qualifier = qualifier.clone();
		if(_reference != null) clone.setReference(_reference);
		clone.setName(getName());
		clone.setColumnName(getColumnName());
		clone.setAttributeIdentifier(this._attributeIdentifier);
		clone.setAggregationClasses(getAggregationClasses());
		return clone;
	}

	/**
	 * Überprüft ob das Attribut atomar ist. Ein nicht-atomares Attribut ist ein Feld ohne Indexangabe und eine Attributliste ohne 'element'. Wird rekursiv
	 * geprüft, d.h. ist das Attribut eine Attributliste mit 'element', so wird überprüft, ob 'element' atomar ist. Ist 'attribute' nicht gesetzt, so wird immer
	 * true zurückgegeben.
	 *
	 * @return True, falls das Attribut atomar ist, false sonst.
	 *
	 * @throws SemanticErrorException Fehler bei der Kommunikation mit der Konfiguration
	 */
	public boolean isAtomic() throws SemanticErrorException {
		if(_reference != null) {
			return _reference.isAtomic();
		}
		else {
			return this._attributeIdentifier.isAtomic();
		}
	}

	/**
	 * Überprüft die Beschreibung des Attributs auf ihre Gültigkeit. Es wird überprüft, ob die Qualifizierung und der Attributname korrekt sind. Die Referenz
	 * ebenfalls überprüft.
	 * @throws CriticalParserException  Exception
	 */
	public void validate() throws CriticalParserException {
		qualifier.valildate();
		if(_attributeIdentifier == null) {
			// das sollte nicht passieren !
			throw new ScopeException(
					InterpreterErrorMessage.INVALID_SCOPE + qualifier.attributeGroup.getPid() + " : " + getName()
			);
		}
		if(_reference != null) {
			if(!((ReferenceAttributeType)_attributeIdentifier.getLastAttribute().getAttributeType()).getReferencedObjectType().equals(_reference.qualifier.object)) {
				throw new CriticalParserException(InterpreterErrorMessage.INVALID_SCOPE);
			}
			_reference.validate();
		}
	}

	/**
	 * Gibt eine Referenz auf dieses Objekt zurück.
	 *
	 * @see de.bsvrz.pua.prot.util.attributes.Elements#getRealElements()
	 */
	@Override
	public HashSet<RealAttributeDescription> getRealElements() {
		final HashSet<RealAttributeDescription> h = new HashSet<RealAttributeDescription>(1);
		h.add(this);
		return h;
	}

	/**
	 * Liefert den Namen des Attributs zurück, der als Spaltenüberschrift verwendet werden soll, wenn kein Spaltenname angegeben wurde.
	 *
	 * @return Spaltenname.
	 */
	public String getPrettyName() {
		return getName();
	}

	/**
	 * Liefert die Qualifizierung des Attributs (inkl. sämtlicher Referenzen) zurück.
	 *
	 * @return Vollständige Qualifzierung. Objekt vom Typ {@link ReferenceQualifier} falls es sich um ein Referenzattribut handelt.
	 */
	public Qualifier getReferenceQualification() {
		if(_reference == null) {
			return qualifier;
		}
		else {
			final ReferenceQualifier qualifier1 = new ReferenceQualifier(this.qualifier, this.getAttributeIdentifier());
			ReferenceQualifier rq = qualifier1;
			RealAttributeDescription rad = this._reference;
			while(rad != null) {
				rq = rq.addReference(rad.qualifier, rad.getAttributeIdentifier());
				rad = rad._reference;
			}
			return qualifier1;
		}
	}

	/** @see de.bsvrz.pua.prot.util.attributes.RealElement#padData() */
	public boolean padData() {
		return _padData;
	}

	public int getIndex() {
		return _index;
	}

	public RealAttributeDescription getReference() {
		return _reference;
	}

	public void setReference(final RealAttributeDescription reference) {

		this._reference = reference;
	}

	public void setAttributeString(final String s, final AttributeGroup attributeGroup) throws SemanticErrorException {
		this.setAttributeIdentifier(new AttributeIdentifier(s, attributeGroup));
	}

	/**
	 * Gibt das letzte Attribut in der Attributkette zurück. Verweist dieses Objekt bspw. auf qKfz.Güte.Index wird hier das Index-Attribut zurückgegeben.
	 * Wird benötigt um den Typ des Attributs zu bestimmen.
	 * @return Attribut
	 */
	public Attribute getAttribute() {
		return this._attributeIdentifier.getLastAttribute();
	}

	public AttributeIdentifier getAttributeIdentifier() {
		return _attributeIdentifier;
	}

	
	public void setAttributeIdentifier(final AttributeIdentifier attributeIdentifier) {
		if(attributeIdentifier == null) throw new NullPointerException();
		_attributeIdentifier = attributeIdentifier;
		setName(_attributeIdentifier.toString());
	}

	public void setIndex(final int index, final String alias) {
		_index = index;
		_alias = alias;
	}



	public String getAlias() {
		return _alias;
	}
}
